# ruff: noqa: S101
from importlib import resources

import pytest

from aana.core.models.image import Image
from aana.core.models.video import Video, VideoParams
from aana.exceptions.io import VideoReadingException
from aana.integrations.external.decord import (
    extract_frames,
    generate_frames,
    get_video_duration,
)


@pytest.mark.parametrize(
    "video_name, extract_fps, fast_mode_enabled, expected_duration, expected_num_frames",
    [
        ("squirrel.mp4", 1.0, False, 10.0, 10),
        ("squirrel.mp4", 0.5, False, 10.0, 5),
        ("squirrel.mp4", 1.0, True, 10.0, 4),
        ("physicsworks_audio.webm", 1.0, False, 203.0, 0),
    ],
)
def test_extract_frames_success(
    video_name, extract_fps, fast_mode_enabled, expected_duration, expected_num_frames
):
    """Test that frames can be extracted from a video successfully."""
    video_path = resources.files("aana.tests.files.videos") / video_name
    video = Video(path=video_path)
    params = VideoParams(extract_fps=extract_fps, fast_mode_enabled=fast_mode_enabled)
    result = extract_frames(video=video, params=params)
    assert "frames" in result
    assert "timestamps" in result
    assert "duration" in result
    assert isinstance(result["duration"], float)
    assert isinstance(result["frames"], list)
    if expected_num_frames == 0:
        assert result["frames"] == []
    else:
        assert isinstance(result["frames"][0], Image)
    assert result["duration"] == expected_duration
    assert len(result["frames"]) == expected_num_frames
    assert len(result["timestamps"]) == expected_num_frames


@pytest.mark.parametrize(
    "video_name, expected_duration",
    [
        ("squirrel.mp4", 10.0),
        ("physicsworks.webm", 203.3),
        ("physicsworks_audio.webm", 203.0),
    ],
)
def test_get_video_duration_success(video_name, expected_duration):
    """Test decord video duration."""
    video_path = resources.files("aana.tests.files.videos") / video_name
    video = Video(path=video_path)
    duration = get_video_duration(video=video)
    assert isinstance(duration, float)
    assert round(duration, 1) == expected_duration


def test_get_video_duration_failure():
    """Test that duration cannot be extracted from a invalid video."""
    # image file instead of video file will create Video object
    # but will fail in get duration function
    path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    with pytest.raises(VideoReadingException):
        invalid_video = Video(path=path)
        get_video_duration(video=invalid_video)


@pytest.mark.parametrize(
    "video_name, extract_fps, fast_mode_enabled, expected_duration, expected_num_frames",
    [
        ("squirrel.mp4", 1.0, False, 10.0, 10),
        ("squirrel.mp4", 0.5, False, 10.0, 5),
        ("squirrel.mp4", 1.0, True, 10.0, 4),
        ("physicsworks_audio.webm", 1.0, False, 203.0, 0),
    ],
)
def test_generate_frames_success(
    video_name, extract_fps, fast_mode_enabled, expected_duration, expected_num_frames
):
    """Test generate_frames.

    generate_frames is a generator function that yields a dictionary
    containing the frames, timestamps and duration of the video.
    """
    video_path = resources.files("aana.tests.files.videos") / video_name
    video = Video(path=video_path)
    params = VideoParams(extract_fps=extract_fps, fast_mode_enabled=fast_mode_enabled)
    gen_frame = generate_frames(video=video, params=params, batch_size=1)
    total_frames = 0
    for result in gen_frame:
        assert "frames" in result
        assert "timestamps" in result
        assert "duration" in result
        assert isinstance(result["duration"], float)
        assert isinstance(result["frames"], list)
        if expected_num_frames == 0:
            assert result["frames"] == []
        else:
            assert isinstance(result["frames"][0], Image)
            assert len(result["frames"]) == 1  # batch_size = 1
            assert len(result["timestamps"]) == 1  # batch_size = 1
            total_frames += 1
        assert result["duration"] == expected_duration

    assert total_frames == expected_num_frames


def test_extract_frames_failure():
    """Test that frames cannot be extracted from a non-existent video."""
    # image file instead of video file will create Video object
    # but will fail in extract_frames
    path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    with pytest.raises(VideoReadingException):
        invalid_video = Video(path=path)
        params = VideoParams(extract_fps=1.0, fast_mode_enabled=False)
        extract_frames(video=invalid_video, params=params)



from aana.core.models.media import MediaId
from aana.exceptions.core import BaseException

__all__ = [
    "NotFoundException",
    "MediaIdAlreadyExistsException",
]


class NotFoundException(BaseException):
    """Raised when an item searched by id is not found."""

    def __init__(self, table_name: str, id: int | MediaId):  # noqa: A002
        """Constructor.

        Args:
            table_name (str): the name of the table being queried.
            id (int | MediaId): the id of the item to be retrieved.
        """
        super().__init__(table=table_name, id=id)
        self.table_name = table_name
        self.id = id
        self.http_status_code = 404

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.table_name, self.id))


class MediaIdAlreadyExistsException(BaseException):
    """Raised when a media_id already exists."""

    def __init__(self, table_name: str, media_id: MediaId):
        """Constructor.

        Args:
            table_name (str): the name of the table being queried.
            media_id (MediaId): the id of the item to be retrieved.
        """
        super().__init__(table=table_name, id=media_id)
        self.table_name = table_name
        self.media_id = media_id

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.table_name, self.media_id))



import base64
from collections.abc import AsyncGenerator
from typing import Any

import torch
from outlines.integrations.vllm import JSONLogitsProcessor, RegexLogitsProcessor
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from ray import serve
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.entrypoints.chat_utils import (
    apply_hf_chat_template,
    apply_mistral_chat_template,
    parse_chat_messages,
)
from vllm.inputs import TokensPrompt
from vllm.model_executor.utils import set_random_seed
from vllm.sampling_params import SamplingParams as VLLMSamplingParams
from vllm.transformers_utils.tokenizer import MistralTokenizer
from vllm.utils import random_uuid

from aana.core.models.base import merged_options, pydantic_protected_fields
from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.custom_config import CustomConfig
from aana.core.models.image import Image
from aana.core.models.image_chat import ImageChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.deployments.base_text_generation_deployment import (
    ChatOutput,
    LLMBatchOutput,
    LLMOutput,
)
from aana.exceptions.runtime import InferenceException, PromptTooLongException
from aana.utils.gpu import get_gpu_memory


class VLLMConfig(BaseModel):
    """The configuration of the vLLM deployment.

    Attributes:
        model_id (str): The model name.
        dtype (Dtype): The data type. Defaults to Dtype.AUTO.
        quantization (str | None): The quantization method. Defaults to None.
        gpu_memory_reserved (float): The GPU memory reserved for the model in MB.
        default_sampling_params (SamplingParams): The default sampling parameters.
            Defaults to SamplingParams(temperature=0, max_tokens=256).
        max_model_len (int | None): The maximum generated text length in tokens. Defaults to None.
        chat_template (str | None): The name of the chat template. If not provided, the chat template
            from the model will be used. Some models may not have a chat template.
            Defaults to None.
        enforce_eager (bool): Whether to enforce eager execution. Defaults to False.
        engine_args (CustomConfig): Extra engine arguments. Defaults to {}.
    """

    model_id: str = Field(validation_alias=AliasChoices("model_id", "model"))
    dtype: Dtype = Field(default=Dtype.AUTO)
    quantization: str | None = Field(default=None)
    gpu_memory_reserved: float
    default_sampling_params: SamplingParams = SamplingParams(
        temperature=0, max_tokens=256
    )
    max_model_len: int | None = Field(default=None)
    chat_template: str | None = Field(default=None)
    enforce_eager: bool = Field(default=False)
    engine_args: CustomConfig = {}

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment
class VLLMDeployment(BaseDeployment):
    """Deployment to serve large language models using vLLM."""

    def __init__(self):
        """Initialize the deployment."""
        super().__init__()
        self.engine = None

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and creates the engine.

        The configuration should contain the following keys:
        - model_id: the model name
        - dtype: the data type (optional, default: "auto")
        - quantization: the quantization method (optional, default: None)
        - gpu_memory_reserved: the GPU memory reserved for the model in mb
        - default_sampling_params: the default sampling parameters.
        - max_model_len: the maximum generated text length in tokens (optional, default: None)
        - chat_template: the name of the chat template (optional, default: None)
        - enforce_eager: whether to enforce eager execution (optional, default: False)
        - engine_args: extra engine arguments (optional, default: {})

        Args:
            config (dict): the configuration of the deployment
        """
        config_obj = VLLMConfig(**config)
        self.model_id = config_obj.model_id
        total_gpu_memory_bytes = get_gpu_memory()
        total_gpu_memory_mb = total_gpu_memory_bytes / 1024**2
        self.gpu_memory_utilization = (
            config_obj.gpu_memory_reserved / total_gpu_memory_mb
        )
        self.default_sampling_params: SamplingParams = (
            config_obj.default_sampling_params
        )
        self.chat_template_name = config_obj.chat_template

        args = AsyncEngineArgs(
            model=config_obj.model_id,
            dtype=config_obj.dtype,
            quantization=config_obj.quantization,
            enforce_eager=config_obj.enforce_eager,
            gpu_memory_utilization=self.gpu_memory_utilization,
            max_model_len=config_obj.max_model_len,
            **config_obj.engine_args,
        )

        # TODO: check if the model is already loaded.
        # If it is and none of the model parameters changed, we don't need to reload the model.

        # create the engine
        self.engine = AsyncLLMEngine.from_engine_args(args)
        self.tokenizer = self.engine.engine.tokenizer.tokenizer
        self.model_config = await self.engine.get_model_config()

    async def check_health(self):
        """Check the health of the deployment and clear torch cache to prevent memory leaks."""
        if self.engine:
            await self.engine.check_health()
            torch.cuda.empty_cache()

        await super().check_health()

    def apply_chat_template(
        self, dialog: ChatDialog | ImageChatDialog
    ) -> tuple[str | list[int], dict | None]:
        """Apply the chat template to the dialog.

        Args:
            dialog (ChatDialog | ImageChatDialog): the dialog (optionally with images)

        Returns:
            tuple[str | list[int], dict | None]: the prompt and the multimodal data
        """

        def image_to_base64(image: Image) -> str:
            """Convert an image to a base64 string."""
            image_data = image.get_content()
            base64_encoded_image = base64.b64encode(image_data)
            base64_string = "data:image;base64," + base64_encoded_image.decode("utf-8")
            return base64_string

        def replace_image_type(messages: list[dict], images: list[Image]) -> list[dict]:
            """Replace the image type with image_url for compatibility with vLLM chat utils.

            vLLM chat utils (parse_chat_messages) only support image_url type for images.
            We need to replace the image type with image_url and provide the actual image content as base64 string.
            """
            i = 0
            for message in messages:
                for item in message["content"]:
                    if item["type"] == "image":
                        item["type"] = "image_url"
                        if i >= len(images):
                            raise ValueError(  # noqa: TRY003
                                "Number of images does not match the number of image items in the message."
                            )
                        item["image_url"] = {"url": image_to_base64(images[i])}
                        i += 1
            if i != len(images):
                raise ValueError(  # noqa: TRY003
                    "Number of images does not match the number of image items in the message."
                )
            return messages

        if isinstance(dialog, ImageChatDialog):
            messages, images = dialog.to_objects()
            messages = replace_image_type(messages, images)
        else:
            messages = dialog.model_dump()["messages"]
            images = None

        conversation, mm_data = parse_chat_messages(
            messages, self.model_config, self.tokenizer
        )

        if isinstance(self.tokenizer, MistralTokenizer):
            prompt = apply_mistral_chat_template(
                self.tokenizer,
                messages=messages,
                add_generation_prompt=True,
            )
        else:
            prompt = apply_hf_chat_template(
                self.tokenizer,
                conversation=conversation,
                chat_template=self.tokenizer.chat_template,
                add_generation_prompt=True,
                # tokenize=True
            )
        return prompt, mm_data

    @exception_handler
    async def generate_stream(  # noqa: C901
        self,
        prompt: str | list[int],
        sampling_params: SamplingParams | None = None,
        mm_data: dict | None = None,
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate completion for the given prompt and stream the results.

        Args:
            prompt (str | list[int]): the prompt or the tokenized prompt
            mm_data (dict | None): the multimodal data
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        if isinstance(prompt, str):
            prompt_token_ids = self.tokenizer.encode(prompt)
        else:
            prompt_token_ids = prompt

        if sampling_params is None:
            sampling_params = self.default_sampling_params
        else:
            sampling_params = merged_options(
                self.default_sampling_params, sampling_params
            )

        json_schema = sampling_params.json_schema
        regex_string = sampling_params.regex_string
        if json_schema is not None:
            logits_processors = [JSONLogitsProcessor(json_schema, self.engine.engine)]
        elif regex_string is not None:
            logits_processors = [RegexLogitsProcessor(regex_string, self.engine.engine)]
        else:
            logits_processors = []

        request_id = None

        if len(prompt_token_ids) > self.model_config.max_model_len:
            raise PromptTooLongException(
                prompt_len=len(prompt_token_ids),
                max_len=self.model_config.max_model_len,
            )

        try:
            # convert SamplingParams to VLLMSamplingParams
            sampling_params_vllm = VLLMSamplingParams(
                **sampling_params.model_dump(
                    exclude_unset=True,
                    exclude=["kwargs", "json_schema", "regex_string"],
                ),
                **sampling_params.kwargs,
                logits_processors=logits_processors,
            )
            # start the request
            request_id = random_uuid()
            # set the random seed for reproducibility
            set_random_seed(42)
            if mm_data is not None:
                inputs = TokensPrompt(
                    prompt_token_ids=prompt_token_ids,
                    multi_modal_data=mm_data,
                )
            else:
                inputs = TokensPrompt(prompt_token_ids=prompt_token_ids)
            results_generator = self.engine.generate(
                inputs,
                sampling_params=sampling_params_vllm,
                request_id=request_id,
            )

            num_returned = 0
            async for request_output in results_generator:
                text_output = request_output.outputs[0].text[num_returned:]
                yield LLMOutput(text=text_output)
                num_returned += len(text_output)
        except GeneratorExit:
            # If the generator is cancelled, we need to cancel the request
            if request_id is not None:
                await self.engine.abort(request_id)
            raise
        except Exception as e:
            raise InferenceException(model_name=self.model_id) from e

    async def generate(
        self,
        prompt: str | list[int],
        sampling_params: SamplingParams | None = None,
        mm_data: dict | None = None,
    ) -> LLMOutput:
        """Generate completion for the given prompt.

        Args:
            prompt (str | list[int]): the prompt or the tokenized prompt
            mm_data (dict | None): the multimodal data
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        generated_text = ""
        async for chunk in self.generate_stream(
            prompt, sampling_params=sampling_params, mm_data=mm_data
        ):
            generated_text += chunk["text"]
        return LLMOutput(text=generated_text)

    async def generate_batch(
        self,
        prompts: list[str] | list[list[int]],
        sampling_params: SamplingParams | None = None,
        mm_data_list: list[dict] | None = None,
    ) -> LLMBatchOutput:
        """Generate completion for the batch of prompts.

        Args:
            prompts (List[str] | List[List[int]]): the list of prompts or the tokenized prompts
            mm_data_list (List[dict] | None): the list of multimodal data
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            LLMBatchOutput: the dictionary with the key "texts"
                            and the list of generated texts as the value
        """
        texts = []
        for i, prompt in enumerate(prompts):
            if mm_data_list is not None:
                text = await self.generate(
                    prompt, sampling_params=sampling_params, mm_data=mm_data_list[i]
                )
            else:
                text = await self.generate(prompt, sampling_params=sampling_params)
            texts.append(text["text"])

        return LLMBatchOutput(texts=texts)

    async def chat(
        self,
        dialog: ChatDialog | ImageChatDialog,
        sampling_params: SamplingParams | None = None,
    ) -> ChatOutput:
        """Chat with the model.

        Args:
            dialog (ChatDialog | ImageChatDialog): the dialog (optionally with images)
            sampling_params (SamplingParams | None): the sampling parameters

        Returns:
            ChatOutput: the dictionary with the key "message"
                        and the response message with a role "assistant"
                        and the generated text as the content
        """
        prompt_ids, mm_data = self.apply_chat_template(dialog)
        response = await self.generate(
            prompt_ids, sampling_params=sampling_params, mm_data=mm_data
        )
        response_message = ChatMessage(content=response["text"], role="assistant")
        return ChatOutput(message=response_message)

    async def chat_stream(
        self,
        dialog: ChatDialog | ImageChatDialog,
        sampling_params: SamplingParams | None = None,
    ) -> AsyncGenerator[LLMOutput, None]:
        """Chat with the model and stream the responses.

        Args:
            dialog (ChatDialog | ImageChatDialog): the dialog (optionally with images)
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the dictionary with the key "text" and the generated text as the value
        """
        prompt_ids, mm_data = self.apply_chat_template(dialog)
        async for chunk in self.generate_stream(
            prompt_ids, sampling_params=sampling_params, mm_data=mm_data
        ):
            yield chunk



import asyncio
import hashlib
import importlib
import random
from pathlib import Path
from typing import Any

from aana.utils.json import jsonify


def get_object_hash(obj: Any) -> str:
    """Get the MD5 hash of an object.

    Objects are converted to JSON strings before hashing.

    Args:
        obj (Any): the object to hash

    Returns:
        str: the MD5 hash of the object
    """
    return hashlib.md5(
        jsonify(obj).encode("utf-8"),
        usedforsecurity=False,
    ).hexdigest()


def import_from(module: str, name, reload=False):
    """Import Module by name."""
    module: Any = __import__(module, fromlist=[str(name)])

    if reload:
        module_path = module.__file__

        # Get the modification timestamp of the module file
        module_last_modified = Path(module_path).stat().st_mtime

        # Check if the module has been imported before
        if hasattr(module, "__last_modified"):
            # Compare the timestamps
            if module_last_modified > module.__last_modified:
                importlib.reload(module)
        else:
            importlib.reload(module)

        # Update the last modified timestamp
        module.__last_modified = module_last_modified

    # return getattr(module, name)
    # getattr doesn't give proper exception, just AttributeError
    # so we use this instead
    return module.__dict__[name]


def import_from_path(path, reload=False):
    """Import a module from path separated by dots.

    :param path: path to the function or class
    :param reload: if True, reload the module
    :return: imported module
    """
    if ":" in path:
        module, name = path.split(":")
    else:
        module, name = ".".join(path.split(".")[:-1]), path.split(".")[-1]
    return import_from(module, name, reload=reload)


def get_module_dir(module_name: str) -> Path:
    """Get the directory of a module.

    Args:
        module_name (str): Name of the module (e.g. 'aana')

    Returns:
        Path: Root directory of the module.
    """
    spec = importlib.util.find_spec(module_name)
    if spec is None:
        raise ImportError(f"Module {module_name} not found")  # noqa: TRY003
    module_path = spec.origin
    return Path(module_path).parent


async def sleep_exponential_backoff(
    initial_delay: float, max_delay: float, attempts: int, jitter: bool = True
):
    """Sleep for an exponentially increasing amount of time with jitter.

    Args:
        initial_delay (float): The initial delay in seconds.
        max_delay (float): The maximum delay in seconds.
        attempts (int): The number of attempts so far.
        jitter (bool): Whether to add jitter to the delay. Default is True.
    """
    # Prevent overflow by using min(attempt, 32) since 2^32 is already huge
    capped_attempt = min(attempts, 32)
    delay = min(initial_delay * (2**capped_attempt), max_delay)
    # Full jitter
    delay_with_jitter = random.uniform(0, delay) if jitter else delay  # noqa: S311
    await asyncio.sleep(delay_with_jitter)



from datetime import datetime, timedelta, timezone
from typing import Any
from uuid import UUID

from sqlalchemy import and_, desc, or_, text
from sqlalchemy.orm import Session

from aana.storage.models.task import Status as TaskStatus
from aana.storage.models.task import TaskEntity
from aana.storage.repository.base import BaseRepository


class TaskRepository(BaseRepository[TaskEntity]):
    """Repository for tasks."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, TaskEntity)

    def read(self, task_id: str | UUID, check: bool = True) -> TaskEntity:
        """Reads a single task by id from the database.

        Args:
            task_id (str | UUID): ID of the task to retrieve
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            The corresponding task from the database if found.

        Raises:
            NotFoundException if the entity is not found and `check` is True.
        """
        if isinstance(task_id, str):
            task_id = UUID(task_id)
        return super().read(task_id, check=check)

    def delete(self, task_id: str | UUID, check: bool = False) -> TaskEntity | None:
        """Deletes a single task by id from the database.

        Args:
            task_id (str | UUID): ID of the task to delete
            check (bool): Whether to check if the task exists before deleting

        Returns:
            The deleted task from the database if found.

        Raises:
            NotFoundException: The id does not correspond to a record in the database.
        """
        if isinstance(task_id, str):
            task_id = UUID(task_id)
        return super().delete(task_id, check)

    def save(self, endpoint: str, data: Any, priority: int = 0):
        """Add a task to the database.

        Args:
            endpoint (str): The endpoint to which the task is assigned.
            data (Any): Data for the task.
            priority (int): Priority of the task (0 is the lowest).

        Returns:
            TaskEntity: The saved task.
        """
        task = TaskEntity(endpoint=endpoint, data=data, priority=priority)
        self.session.add(task)
        self.session.commit()
        return task

    def fetch_unprocessed_tasks(
        self,
        limit: int | None = None,
        max_retries: int = 1,
        retryable_exceptions: list[str] | None = None,
    ) -> list[TaskEntity]:
        """Fetches unprocessed tasks and marks them as ASSIGNED.

        The task is considered unprocessed if it is in CREATED or NOT_FINISHED state.

        The function runs in a transaction and locks the rows to prevent race condition
        if multiple task queue deployments are running concurrently.

        IMPORTANT: The lock doesn't work with SQLite. If you are using SQLite, you should
        only run one task queue deployment at a time. Otherwise, you may encounter
        race conditions.

        Args:
            limit (int | None): The maximum number of tasks to fetch. If None, fetch all.
            max_retries (int): The maximum number of retries for a task.
            retryable_exceptions (list[str] | None): The list of exceptions that should be retried.

        Returns:
            list[TaskEntity]: the unprocessed tasks.
        """
        if retryable_exceptions:
            # Convert the list of exceptions to a string for the query:
            # e.g., ["InferenceException", "ValueError"] -> "'InferenceException', 'ValueError'"
            exceptions_str = ", ".join([f"'{ex}'" for ex in retryable_exceptions])
            if self.session.bind.dialect.name == "postgresql":
                exception_name_query = f"result->>'error' IN ({exceptions_str})"
            elif self.session.bind.dialect.name == "sqlite":
                exception_name_query = (
                    f"json_extract(result, '$.error') IN ({exceptions_str})"
                )

            tasks = (
                self.session.query(TaskEntity)
                .filter(
                    or_(
                        TaskEntity.status.in_(
                            [TaskStatus.CREATED, TaskStatus.NOT_FINISHED]
                        ),
                        and_(
                            TaskEntity.status == TaskStatus.FAILED,
                            text(exception_name_query),
                            TaskEntity.num_retries < max_retries,
                        ),
                    )
                )
                .order_by(desc(TaskEntity.priority), TaskEntity.created_at)
                .limit(limit)
                .populate_existing()
                .with_for_update(skip_locked=True)
                .all()
            )
        else:
            tasks = (
                self.session.query(TaskEntity)
                .filter(
                    TaskEntity.status.in_(
                        [TaskStatus.CREATED, TaskStatus.NOT_FINISHED]
                    ),
                )
                .order_by(desc(TaskEntity.priority), TaskEntity.created_at)
                .limit(limit)
                .populate_existing()
                .with_for_update(skip_locked=True)
                .all()
            )
        for task in tasks:
            self.update_status(
                task_id=task.id,
                status=TaskStatus.ASSIGNED,
                progress=0,
                commit=False,
            )
        self.session.commit()
        return tasks

    def update_status(
        self,
        task_id: str,
        status: TaskStatus,
        progress: int | None = None,
        result: Any = None,
        commit: bool = True,
    ):
        """Update the status of a task.

        Args:
            task_id (str): The ID of the task.
            status (TaskStatus): The new status.
            progress (int | None): The progress. If None, the progress will not be updated.
            result (Any): The result.
            commit (bool): Whether to commit the transaction.
        """
        task = self.read(task_id)
        if status == TaskStatus.COMPLETED or status == TaskStatus.FAILED:
            task.completed_at = datetime.now(timezone.utc)
        if status == TaskStatus.ASSIGNED:
            task.assigned_at = datetime.now(timezone.utc)
            task.num_retries += 1
        if progress is not None:
            task.progress = progress
        task.status = status
        task.result = result
        if commit:
            self.session.commit()

    def get_active_tasks(self) -> list[TaskEntity]:
        """Fetches all active tasks.

        The task is considered active if it is in RUNNING or ASSIGNED state.

        Returns:
            list[TaskEntity]: the active tasks.
        """
        tasks = (
            self.session.query(TaskEntity)
            .filter(TaskEntity.status.in_([TaskStatus.RUNNING, TaskStatus.ASSIGNED]))
            .all()
        )
        return tasks

    def filter_incomplete_tasks(self, task_ids: list[str]) -> list[str]:
        """Remove the task IDs that are already completed (COMPLETED or FAILED).

        Args:
            task_ids (list[str]): The task IDs to filter.

        Returns:
            list[str]: The task IDs that are not completed.
        """
        task_ids = [UUID(task_id) for task_id in task_ids]
        tasks = (
            self.session.query(TaskEntity)
            .filter(
                and_(
                    TaskEntity.id.in_(task_ids),
                    TaskEntity.status.not_in(
                        [
                            TaskStatus.COMPLETED,
                            TaskStatus.FAILED,
                            TaskStatus.NOT_FINISHED,
                        ]
                    ),
                )
            )
            .all()
        )
        incomplete_task_ids = [str(task.id) for task in tasks]
        return incomplete_task_ids

    def update_expired_tasks(
        self, execution_timeout: float, heartbeat_timeout: float, max_retries: int
    ) -> list[TaskEntity]:
        """Fetches all tasks that are expired and updates their status.

        The task is considered expired if it is in RUNNING or ASSIGNED state and the
        updated_at time is older than the execution_timeout.

        If the task has exceeded the maximum number of retries, it will be marked as FAILED.
        If the task has not exceeded the maximum number of retries, it will be marked as NOT_FINISHED and
        be retried again.

        The function runs in a transaction and locks the rows to prevent race condition
        if multiple task queue deployments are running concurrently.

        IMPORTANT: The lock doesn't work with SQLite. If you are using SQLite, you should
        only run one task queue deployment at a time. Otherwise, you may encounter
        race conditions.

        Args:
            execution_timeout (float): The maximum execution time for a task in seconds
            heartbeat_timeout (float): The maximum time since the last heartbeat in seconds
            max_retries (int): The maximum number of retries for a task

        Returns:
            list[TaskEntity]: the expired tasks.
        """
        timeout_cutoff = datetime.now(timezone.utc) - timedelta(
            seconds=execution_timeout
        )
        heartbeat_cutoff = datetime.now(timezone.utc) - timedelta(
            seconds=heartbeat_timeout
        )
        tasks = (
            self.session.query(TaskEntity)
            .filter(
                and_(
                    TaskEntity.status.in_([TaskStatus.RUNNING, TaskStatus.ASSIGNED]),
                    or_(
                        TaskEntity.assigned_at <= timeout_cutoff,
                        TaskEntity.updated_at <= heartbeat_cutoff,
                    ),
                ),
            )
            .populate_existing()
            .with_for_update(skip_locked=True)
            .all()
        )
        for task in tasks:
            if task.num_retries >= max_retries:
                if task.assigned_at.astimezone(timezone.utc) <= timeout_cutoff:
                    result = {
                        "error": "TimeoutError",
                        "message": (
                            f"Task execution timed out after {execution_timeout} seconds and "
                            f"exceeded the maximum number of retries ({max_retries})"
                        ),
                    }
                else:
                    result = {
                        "error": "HeartbeatTimeoutError",
                        "message": (
                            f"The task has not received a heartbeat for {heartbeat_timeout} seconds and "
                            f"exceeded the maximum number of retries ({max_retries})"
                        ),
                    }

                self.update_status(
                    task_id=task.id,
                    status=TaskStatus.FAILED,
                    progress=0,
                    result=result,
                    commit=False,
                )
            else:
                self.update_status(
                    task_id=task.id,
                    status=TaskStatus.NOT_FINISHED,
                    progress=0,
                    commit=False,
                )
        self.session.commit()
        return tasks

    def heartbeat(self, task_ids: list[str] | set[str]):
        """Updates the updated_at timestamp for multiple tasks.

        Args:
            task_ids (list[str] | set[str]): List or set of task IDs to update
        """
        task_ids = [
            UUID(task_id) if isinstance(task_id, str) else task_id
            for task_id in task_ids
        ]
        self.session.query(TaskEntity).filter(TaskEntity.id.in_(task_ids)).update(
            {TaskEntity.updated_at: datetime.now(timezone.utc)},
            synchronize_session=False,
        )
        self.session.commit()



# ruff: noqa: S101
import hashlib
from importlib import resources
from pathlib import Path

import pytest

from aana.configs.settings import settings
from aana.core.models.video import Video, VideoInput, VideoMetadata
from aana.exceptions.io import DownloadException, VideoReadingException
from aana.integrations.external.yt_dlp import download_video, get_video_metadata


def test_video():
    """Test that the video can be created from path, url, or content."""
    # Test creation from a path
    try:
        path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
        video = Video(path=path, save_on_disk=False)
        assert video.path == path
        assert video.content is None
        assert video.url is None

        content = video.get_content()
        assert len(content) > 0
    finally:
        video.cleanup()

    # Test creation from a URL
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video = Video(url=url, save_on_disk=False)
        assert video.path is None
        assert video.content is None
        assert video.url == url

        content = video.get_content()
        assert len(content) > 0
    finally:
        video.cleanup()

    # Test creation from content
    try:
        path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
        content = path.read_bytes()
        video = Video(content=content, save_on_disk=False)
        assert video.path is None
        assert video.content == content
        assert video.url is None

        assert video.get_content() == content
    finally:
        video.cleanup()


def test_media_dir():
    """Test that the media_dir is set correctly."""
    # Test saving from URL to disk
    video_dir = settings.video_dir
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video = Video(url=url, save_on_disk=True)
        assert video.media_dir == video_dir
        assert video.content is None
        assert video.url == url
        assert video.path.exists()
        assert str(video.path).startswith(str(video_dir))
    finally:
        video.cleanup()


def test_video_path_not_exist():
    """Test that the video can't be created from path if the path doesn't exist."""
    path = Path("path/to/video_that_does_not_exist.mp4")
    with pytest.raises(FileNotFoundError):
        Video(path=path)


def test_save_video():
    """Test that save_on_disk works for video."""
    # Test that the video is saved to disk when save_on_disk is True
    try:
        path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
        video = Video(path=path, save_on_disk=True)
        assert video.path == path
        assert video.content is None
        assert video.url is None
        assert video.path.exists()
    finally:
        video.cleanup()
    assert video.path.exists()  # Cleanup should NOT delete the file if path is provided

    # Test saving from URL to disk
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video = Video(url=url, save_on_disk=True)
        assert video.content is None
        assert video.url == url
        assert video.path.exists()
    finally:
        video.cleanup()

    # Test saving from content to disk
    try:
        path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
        content = path.read_bytes()
        video = Video(content=content, save_on_disk=True)
        assert video.content == content
        assert video.url is None
        assert video.path.exists()
    finally:
        video.cleanup()


def test_cleanup():
    """Test that cleanup works for video."""
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video = Video(url=url, save_on_disk=True)
        assert video.path.exists()
    finally:
        video.cleanup()
        assert not video.path.exists()

    try:
        path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
        video = Video(path=path, save_on_disk=True)
        assert video.path.exists()
    finally:
        video.cleanup()
        assert (
            video.path.exists()
        )  # Cleanup should NOT delete the file if path is provided


def test_at_least_one_input():
    """Test that at least one input is provided for video."""
    with pytest.raises(ValueError):
        Video(save_on_disk=False)

    with pytest.raises(ValueError):
        Video(save_on_disk=True)


def test_download_video():
    """Test download_video."""
    # Test VideoInput with path
    path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
    video_input = VideoInput(path=str(path))
    video = download_video(video_input)
    assert isinstance(video, Video)
    assert video.path == path
    assert video.content is None
    assert video.url is None

    # Test VideoInput with url
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video_input = VideoInput(url=url)
        video = download_video(video_input)
        assert isinstance(video, Video)
        assert video.path is not None
        assert video.content is None
        assert video.url == url
        assert video.path.exists()
        assert video.media_id == video_input.media_id
    finally:
        video.cleanup()

    # Test url that doesn't contain a video
    url = "https://mobius-public.s3.eu-west-1.amazonaws.com/Starry_Night.jpeg"
    video_input = VideoInput(url=url)
    with pytest.raises(VideoReadingException):
        download_video(video_input)

    # Test Video object as input
    path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
    video = Video(path=path)
    downloaded_video = download_video(video)
    assert downloaded_video == video


@pytest.mark.parametrize(
    "url, title, description, duration",
    [
        (
            "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
            "Rick Astley - Never Gonna Give You Up (Official Music Video)",
            "The official video for “Never Gonna Give You Up” by Rick Astley.",
            212.0,
        ),
        (
            "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4",
            "squirrel",
            "",
            None,
        ),
    ],
)
def test_get_video_metadata_success(mocker, url, title, description, duration):
    """Test getting video metadata."""
    # Mock the yt-dlp YoutubeDL class and its methods
    mock_ydl = mocker.patch("yt_dlp.YoutubeDL", autospec=True)
    mock_ydl_instance = mock_ydl.return_value
    mock_ydl_instance.__enter__.return_value = mock_ydl_instance

    mock_info = {
        "title": title,
        "description": description,
        "duration": duration,
    }
    mock_ydl_instance.extract_info.return_value = mock_info

    metadata = get_video_metadata(url)
    assert isinstance(metadata, VideoMetadata)
    assert metadata.title == title
    assert metadata.description.startswith(description)
    assert metadata.duration == duration


def test_get_video_metadata_failure():
    """Test that get_video_metadata fails for invalid URLs."""
    url = "https://www.youtube.com/watch?v=invalid_url"
    with pytest.raises(DownloadException):
        get_video_metadata(url)


def test_download_youtube_video(mocker):
    """Test downloading a YouTube video."""
    youtube_url = "https://www.youtube.com/watch?v=dQw4w9WgXcQ"
    youtube_url_hash = hashlib.md5(
        youtube_url.encode(), usedforsecurity=False
    ).hexdigest()
    video_dir = settings.video_dir
    expected_path = Path(video_dir) / f"{youtube_url_hash}.mp4"

    # Mock the VideoInput object
    youtube_video_input = VideoInput(url=youtube_url, media_id="dQw4w9WgXcQ")

    # Mock the yt-dlp YoutubeDL class and its methods
    mock_ydl = mocker.patch("yt_dlp.YoutubeDL", autospec=True)
    mock_ydl_instance = mock_ydl.return_value
    mock_ydl_instance.__enter__.return_value = mock_ydl_instance

    mock_info = {
        "title": "Rick Astley - Never Gonna Give You Up (Official Music Video)",
        "description": "The official video for “Never Gonna Give You Up” by Rick Astley.",
        "ext": "mp4",
    }
    mock_ydl_instance.extract_info.return_value = mock_info
    mock_ydl_instance.prepare_filename.return_value = str(expected_path)
    mock_ydl_instance.download.return_value = None

    # # Mock Path.exists to simulate that the file will exist after download
    mocker.patch("aana.integrations.external.yt_dlp.Path.exists", return_value=True)

    # Mock decord.VideoReader to avoid trying to read the non-existent mocked file
    mocker.patch("decord.VideoReader", autospec=True)

    video = download_video(youtube_video_input)

    assert isinstance(video, Video)
    assert video.path.with_suffix("") == expected_path.with_suffix("")
    assert video.path is not None
    assert video.path.exists()
    assert video.content is None
    assert video.url == youtube_url
    assert video.media_id == "dQw4w9WgXcQ"
    assert video.title == "Rick Astley - Never Gonna Give You Up (Official Music Video)"
    assert video.description.startswith(
        "The official video for “Never Gonna Give You Up” by Rick Astley."
    )


def test_download_youtube_video_failure():
    """Test YoutubeVideoInput with invalid youtube URL."""
    youtube_url = "https://www.youtube.com/watch?v=invalid_url"
    youtube_video_input = VideoInput(url=youtube_url)
    with pytest.raises(DownloadException):
        download_video(youtube_video_input)



import hashlib
from pathlib import Path


def get_sha256_hash_file(filename: Path) -> str:
    """Compute SHA-256 hash of a file without loading it entirely in memory.

    Args:
        filename (Path): Path to the file to be hashed.

    Returns:
        str: SHA-256 hash of the file in hexadecimal format.
    """
    # Create a sha256 hash object
    sha256 = hashlib.sha256()

    # Open the file in binary mode
    with Path.open(filename, "rb") as f:
        # Read and update hash in chunks of 4K
        for chunk in iter(lambda: f.read(4096), b""):
            sha256.update(chunk)

    # Return the hexadecimal representation of the digest
    return sha256.hexdigest()



from __future__ import annotations  # Let classes use themselves in type annotations

import typing

from sqlalchemy import CheckConstraint
from sqlalchemy.orm import Mapped, mapped_column

from aana.storage.models.base import BaseEntity, TimeStampEntity

if typing.TYPE_CHECKING:
    from aana.core.models.captions import Caption


class CaptionEntity(BaseEntity, TimeStampEntity):
    """ORM model for video captions.

    Attributes:
        id (int): Unique identifier for the caption.
        model (str): Name of the model used to generate the caption.
        frame_id (int): The 0-based frame id of video for caption.
        caption (str): Frame caption.
        timestamp (float): Frame timestamp in seconds.
        caption_type (str): The type of caption (populated automatically by ORM based on `polymorphic_identity` of subclass).
    """

    __tablename__ = "caption"

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True)
    model: Mapped[str] = mapped_column(
        nullable=False, comment="Name of model used to generate the caption"
    )
    frame_id: Mapped[int] = mapped_column(
        CheckConstraint("frame_id >= 0", "frame_id_positive"),
        comment="The 0-based frame id of video for caption",
    )
    caption: Mapped[str] = mapped_column(comment="Frame caption")
    timestamp: Mapped[float] = mapped_column(
        CheckConstraint("timestamp >= 0", name="timestamp_positive"),
        comment="Frame timestamp in seconds",
    )
    caption_type: Mapped[str] = mapped_column(comment="The type of caption")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "caption",
        "polymorphic_on": "caption_type",
    }

    @classmethod
    def from_caption_output(
        cls,
        model_name: str,
        caption: Caption,
        frame_id: int,
        timestamp: float,
    ) -> CaptionEntity:
        """Converts a Caption pydantic model to a CaptionEntity.

        Args:
            model_name (str): Name of the model used to generate the caption.
            caption (Caption): Caption pydantic model.
            frame_id (int): The 0-based frame id of video for caption.
            timestamp (float): Frame timestamp in seconds.

        Returns:
            CaptionEntity: ORM model for video captions.
        """
        return CaptionEntity(
            model=model_name,
            frame_id=frame_id,
            caption=str(caption),
            timestamp=timestamp,
        )



# ruff: noqa: S101

from importlib import resources

import pytest

from aana.core.models.image import Image
from aana.core.models.image_chat import (
    ImageChatDialog,
    ImageChatMessage,
    ImageContent,
    TextContent,
)


@pytest.fixture
def input_image():
    """Gets an Image for test."""
    image_path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    return Image(path=image_path)


@pytest.fixture
def input_prompt():
    """Gets a text for text."""
    return "describe images with more info!"


def test_text_content_creation(input_image, input_prompt):
    """Test that a text content can be created."""
    content = TextContent(text=input_prompt)
    assert content.type == "text"
    assert content.text == input_prompt

    # Value error if we tried to init TextContent with different type attr
    with pytest.raises(ValueError):
        content = TextContent(text=input_prompt, type="image")

    # Value error if we tried to set text field with non-str value
    with pytest.raises(ValueError):
        content = TextContent(text=input_image, type="text")


def test_image_content_creation(input_image):
    """Test that a image content can be created."""
    content = ImageContent(image=input_image)
    assert content.type == "image"
    assert content.image.path == input_image.path

    # Value error if we tried to init ImageContent with different type attr
    with pytest.raises(ValueError):
        content = ImageContent(image=input_image, type="text")

    # Value error if we tried to set image field with non-image value
    with pytest.raises(ValueError):
        content = ImageContent(image="Image")


def test_image_dialog_creation(input_image, input_prompt):
    """Test that a image chat dialog can be created."""
    system_messages = ImageChatMessage(
        content=[ImageContent(image=input_image), TextContent(text=input_prompt)],
        role="system",
    )

    user_messages = ImageChatMessage(
        content=[ImageContent(image=input_image), TextContent(text=input_prompt)],
        role="user",
    )

    dialog = ImageChatDialog(messages=[system_messages, user_messages])

    assert len(dialog.messages) == 2

    for message, role in zip(dialog.messages, ["system", "user"], strict=False):
        assert isinstance(message, ImageChatMessage)
        assert message.role == role
        assert len(message.content) == 2

        assert isinstance(message.content[0], ImageContent)
        assert message.content[0].image.path == input_image.path

        assert isinstance(message.content[1], TextContent)
        assert message.content[1].text == input_prompt

    messages, images = dialog.to_objects()
    assert len(messages) == 2
    assert len(images) == 2

    for message, role in zip(messages, ["system", "user"], strict=False):
        assert message["role"] == role
        assert len(message["content"]) == 2
        assert message["content"][0] == {"type": "image"}

        assert message["content"][1] == {"type": "text", "text": input_prompt}


def test_image_dialog_creation_from_prompt(input_image, input_prompt):
    """Test that a image chat dialog can be created from prompt."""
    image_list = [input_image for _ in range(5)]

    dialog = ImageChatDialog.from_prompt(input_prompt, images=image_list)

    assert len(dialog.messages) == 1
    message = dialog.messages[0]

    assert isinstance(message, ImageChatMessage)
    assert message.role == "user"
    assert len(message.content) == 6

    for i in range(5):
        assert isinstance(message.content[i], ImageContent)
        assert message.content[i].image.path == input_image.path

    assert isinstance(message.content[5], TextContent)
    assert message.content[5].text == input_prompt

    messages, images = dialog.to_objects()
    assert len(messages) == 1
    assert len(images) == 5

    message = messages[0]
    assert len(message["content"]) == 6
    for i in range(5):
        assert message["content"][i] == {"type": "image"}

    assert message["content"][5] == {"type": "text", "text": input_prompt}


def test_image_dialog_creation_from_list(input_image, input_prompt):
    """Test that a image chat dialog can be created from prompt."""
    messages = [
        {
            "content": [
                {"type": "image", "image": input_image},
                {"type": "text", "text": input_prompt},
            ],
            "role": "system",
        },
        {
            "content": [
                {"type": "image", "image": input_image},
                {"type": "text", "text": input_prompt},
            ],
            "role": "user",
        },
    ]

    dialog = ImageChatDialog.from_list(messages)

    assert len(dialog.messages) == 2

    for message, role in zip(dialog.messages, ["system", "user"], strict=False):
        assert isinstance(message, ImageChatMessage)
        assert message.role == role
        assert len(message.content) == 2

        assert isinstance(message.content[0], ImageContent)
        assert message.content[0].image.path == input_image.path

        assert isinstance(message.content[1], TextContent)
        assert message.content[1].text == input_prompt

    messages, images = dialog.to_objects()
    assert len(messages) == 2
    assert len(images) == 2

    for message, role in zip(messages, ["system", "user"], strict=False):
        assert message["role"] == role
        assert len(message["content"]) == 2
        assert message["content"][0] == {"type": "image"}

        assert message["content"][1] == {"type": "text", "text": input_prompt}



# ruff: noqa: S101
import json
from collections import defaultdict
from importlib import resources
from pathlib import Path

import pytest

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_to_dict
from aana.core.models.vad import VadSegment
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.whisper_deployment import (
    WhisperComputeType,
    WhisperConfig,
    WhisperDeployment,
    WhisperModelSize,
)
from aana.tests.utils import verify_deployment_results

deployments = [
    (
        "whisper_tiny",
        WhisperDeployment.options(
            num_replicas=1,
            user_config=WhisperConfig(
                model_size=WhisperModelSize.TINY,
                compute_type=WhisperComputeType.FLOAT32,
            ).model_dump(mode="json"),
        ),
    ),
    (
        "whisper_medium",
        WhisperDeployment.options(
            num_replicas=1,
            ray_actor_options={"num_gpus": 0.25},
            user_config=WhisperConfig(
                model_size=WhisperModelSize.MEDIUM,
                compute_type=WhisperComputeType.FLOAT16,
            ).model_dump(mode="json"),
        ),
    ),
    (
        "whisper_turbo",
        WhisperDeployment.options(
            num_replicas=1,
            ray_actor_options={"num_gpus": 0.25},
            user_config=WhisperConfig(
                model_size=WhisperModelSize.TURBO,
                compute_type=WhisperComputeType.FLOAT16,
            ).model_dump(mode="json"),
        ),
    ),
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestWhisperDeployment:
    """Test Whisper deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("audio_file", ["squirrel.wav", "physicsworks.wav"])
    async def test_transcribe(self, setup_deployment, audio_file):
        """Test transcribe methods."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "whisper"
            / f"{deployment_name}_{audio_file}.json"
        )

        # Test transcribe method
        path = resources.files("aana.tests.files.audios") / audio_file
        assert path.exists(), f"Audio not found: {path}"
        audio = Audio(path=path, media_id=audio_file)

        output = await handle.transcribe(
            audio=audio, params=WhisperParams(word_timestamps=True, temperature=0.0)
        )

        verify_deployment_results(expected_output_path, output)

        # Test transcribe_stream method
        stream = handle.transcribe_stream(
            audio=audio, params=WhisperParams(word_timestamps=True, temperature=0.0)
        )

        # Combine individual segments and compare with the final dict
        grouped_dict = defaultdict(list)
        transcript = ""
        async for chunk in stream:
            output = pydantic_to_dict(chunk)
            transcript += output["transcription"]["text"]
            grouped_dict["segments"].extend(output.get("segments", []))

        grouped_dict["transcription"] = {"text": transcript}
        grouped_dict["transcription_info"] = output.get("transcription_info")

        verify_deployment_results(expected_output_path, grouped_dict)

        # Test transcribe_in_chunks method: Note that the expected asr output is different
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "whisper"
            / f"{deployment_name}_{audio_file}_batched.json"
        )
        audio_file_name = audio_file.removesuffix(".wav")
        # Get expected vad segments
        vad_path = (
            resources.files("aana.tests.files.expected.vad") / f"{audio_file_name}.json"
        )

        assert vad_path.exists(), f"vad expected predictions not found: {vad_path}"
        with Path(vad_path) as path, path.open() as f:
            expected_output_vad = json.load(f)
        vad_input = [
            VadSegment(time_interval=seg["time_interval"], segments=seg["segments"])
            for seg in expected_output_vad["segments"]
        ]

        batched_stream = handle.transcribe_in_chunks(
            audio=audio,
            vad_segments=vad_input,
            batch_size=16,
            params=BatchedWhisperParams(temperature=0.0),
        )
        # Combine individual segments and compare with the final dict
        grouped_dict = defaultdict(list)
        transcript = ""
        async for chunk in batched_stream:
            output = pydantic_to_dict(chunk)
            transcript += output["transcription"]["text"]
            grouped_dict["segments"].extend(output.get("segments", []))
        grouped_dict["transcription"] = {"text": transcript}
        grouped_dict["transcription_info"] = output.get("transcription_info")
        verify_deployment_results(expected_output_path, grouped_dict)

        # Test with whisper internal VAD
        batched_stream = handle.transcribe_in_chunks(
            audio=audio,
            batch_size=16,
            params=BatchedWhisperParams(temperature=0.0),
        )
        # Combine individual segments and compare with the final dict
        grouped_dict = defaultdict(list)
        transcript = ""
        async for chunk in batched_stream:
            output = pydantic_to_dict(chunk)
            transcript += output["transcription"]["text"]
            grouped_dict["segments"].extend(output.get("segments", []))
        grouped_dict["transcription"] = {"text": transcript}
        grouped_dict["transcription_info"] = output.get("transcription_info")
        verify_deployment_results(expected_output_path, grouped_dict)



from sqlalchemy.orm import Session, sessionmaker

from aana.storage.engine import engine

__all__ = ["get_session"]

SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_session() -> Session:
    """Get a new SQLAlchemy Session object.

    Returns:
        Session: SQLAlchemy Session object.
    """
    return SessionLocal()



# ruff: noqa: S101
from importlib import resources

import pytest

from aana.core.models.image import Image
from aana.core.models.types import Dtype
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.hf_blip2_deployment import HFBlip2Config, HFBlip2Deployment
from aana.tests.utils import verify_deployment_results

deployments = [
    (
        "blip2_deployment",
        HFBlip2Deployment.options(
            num_replicas=1,
            max_ongoing_requests=1000,
            ray_actor_options={"num_gpus": 0.25},
            user_config=HFBlip2Config(
                model="Salesforce/blip2-opt-2.7b",
                dtype=Dtype.FLOAT16,
                batch_size=2,
                num_processing_threads=2,
            ).model_dump(mode="json"),
        ),
    )
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestHFBlip2Deployment:
    """Test HuggingFace BLIP2 deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("image_name", ["Starry_Night.jpeg"])
    async def test_methods(self, setup_deployment, image_name):
        """Test HuggingFace BLIP2 methods."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "hf_blip2"
            / f"{deployment_name}_{image_name}.json"
        )

        path = resources.files("aana.tests.files.images") / image_name
        image = Image(path=path, save_on_disk=False, media_id=image_name)

        output = await handle.generate(image=image)
        caption = output["caption"]
        verify_deployment_results(expected_output_path, caption)

        images = [image] * 8

        output = await handle.generate_batch(images=images)
        captions = output["captions"]

        assert len(captions) == 8
        for caption in captions:
            verify_deployment_results(expected_output_path, caption)



from time import monotonic

from typing_extensions import override

from aana.api.event_handlers.event_handler import EventHandler
from aana.exceptions.runtime import TooManyRequestsException


class RateLimitHandler(EventHandler):
    """Event handler that raises TooManyRequestsException if the rate limit is exceeded.

    Attributes:
        capacity (int): number of resources (requests) per interval
        rate (float): the interval for the limit in seconds
    """

    capacity: int
    interval: float

    _calls: list

    def __init__(self, capacity: int, interval: float):
        """Constructor."""
        self.capacity = capacity
        self.interval = interval
        self._calls = []

    def _clear_expired(self, expired: float):
        """Removes expired items from list of resources.

        Arguments:
            expired: timestamp before which to clear, as output from time.monotonic()
        """
        while self._calls and self._calls[0] < expired:
            self._calls.pop(0)

    def _acquire(self):
        """Checks if we can acquire (process) a resource.

        Raises:
            TooManyRequestsException: if the rate limit has been reached
        """
        now = monotonic()
        expired = now - self.interval
        self._clear_expired(expired)
        if len(self._calls) < self.capacity:
            self._calls.append(now)
        else:
            raise TooManyRequestsException(self.capacity, self.interval)

    @override
    def handle(self, event_name: str, *args, **kwargs):
        """Handle the event by checking against rate limiting parameters.

        Arguments:
            event_name (str): the name of the event to handle
            *args (list): args for the event
            **kwargs (dict): keyword args for the event

        Raises:
            TooManyRequestsException: if the rate limit has been reached
        """
        # if the endpoint execution is deferred, we don't want to rate limit it
        defer = kwargs.get("defer", False)
        if not defer:
            self._acquire()



from enum import Enum

from pydantic import BaseModel, ConfigDict, Field
from ray.serve.schema import ApplicationStatus


class SDKStatus(str, Enum):
    """The status of the SDK."""

    UNHEALTHY = "UNHEALTHY"
    RUNNING = "RUNNING"
    DEPLOYING = "DEPLOYING"


class DeploymentStatus(BaseModel):
    """The status of a deployment."""

    status: ApplicationStatus = Field(description="The status of the deployment.")
    message: str = Field(
        description="The message for more information like error message."
    )


class SDKStatusResponse(BaseModel):
    """The response for the SDK status endpoint.

    Attributes:
        status (SDKStatus): The status of the SDK.
        message (str): The message for more information like error message.
        deployments (dict[str, DeploymentStatus]): The status of each deployment in the Aana app.
    """

    status: SDKStatus = Field(description="The status of the SDK.")
    message: str = Field(
        description="The message for more information like error message."
    )
    deployments: dict[str, DeploymentStatus] = Field(
        description="The status of each deployment in the Aana app."
    )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "The response for the SDK status endpoint.",
            "examples": [
                {
                    "status": "RUNNING",
                    "message": "",
                    "deployments": {
                        "app": {
                            "status": "RUNNING",
                            "message": "",
                        },
                        "lowercase_deployment": {
                            "status": "RUNNING",
                            "message": "",
                        },
                    },
                },
                {
                    "status": "UNHEALTHY",
                    "message": "Error: Lowercase (lowercase_deployment): A replica's health check failed. "
                    "This deployment will be UNHEALTHY until the replica recovers or a new deploy happens.",
                    "deployments": {
                        "app": {
                            "status": "RUNNING",
                            "message": "",
                        },
                        "lowercase_deployment": {
                            "status": "UNHEALTHY",
                            "message": "A replica's health check failed. This deployment will be UNHEALTHY "
                            "until the replica recovers or a new deploy happens.",
                        },
                    },
                },
            ],
        }
    )



from typing import TypeVar

from sqlalchemy.orm import Session

from aana.core.models.media import MediaId
from aana.core.models.video import Video, VideoMetadata
from aana.storage.models import VideoEntity
from aana.storage.repository.media import MediaRepository

V = TypeVar("V", bound=VideoEntity)


class VideoRepository(MediaRepository[V]):
    """Repository for videos."""

    def __init__(self, session: Session, model_class: type[V] = VideoEntity):
        """Constructor."""
        super().__init__(session, model_class)

    def save(self, video: Video) -> dict:
        """Saves a video to datastore.

        Args:
            video (Video): The video object.

        Returns:
            dict: The dictionary with media ID.
        """
        video_entity = VideoEntity(
            id=video.media_id,
            path=str(video.path),
            url=video.url,
            title=video.title,
            description=video.description,
        )

        self.create(video_entity)
        return {
            "media_id": video_entity.id,
        }

    def get_metadata(self, media_id: MediaId) -> VideoMetadata:
        """Get the metadata of a video.

        Args:
            media_id (MediaId): The media ID.

        Returns:
            VideoMetadata: The video metadata.
        """
        entity: VideoEntity = self.read(media_id)
        return VideoMetadata(title=entity.title, description=entity.description)



# ruff: noqa: S101
from typing import Any, TypedDict

import haystack
import haystack.components.preprocessors
import numpy as np
import pytest
from haystack import Document
from haystack.components.preprocessors import TextCleaner
from ray import serve

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment
from aana.integrations.haystack.deployment_component import AanaDeploymentComponent


class DummyEmbedderOutput(TypedDict):
    """Output of the dummy embedder."""

    embeddings: np.ndarray


@serve.deployment
class DummyEmbedder(BaseDeployment):
    """Simple deployment to store video data."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration to the deployment and initialize it."""
        pass

    async def embed(self, texts: list[str]) -> DummyEmbedderOutput:
        """Generate a dummy embedding for the texts."""
        return {"embeddings": np.random.default_rng().random((len(texts), 8))}

    async def embed_fail(self, texts: list[str]) -> DummyEmbedderOutput:
        """Fail to generate a dummy embedding for the texts."""
        raise Exception("Dummy exception")  # noqa: TRY002, TRY003


@pytest.fixture(scope="module")
def setup_dummy_embedder_deployment(create_app):
    """Set up the dummy embedder deployment."""
    deployment_name = "dummy_embedder"
    deployment = DummyEmbedder.options(num_replicas=1, user_config={})
    deployments = [
        {
            "name": deployment_name,
            "instance": deployment,
        }
    ]
    endpoints = []

    return deployment_name, create_app(deployments, endpoints)


@pytest.mark.asyncio
async def test_haystack_wrapper(setup_dummy_embedder_deployment):
    """Tests haystack wrapper for deployments."""
    deployment_name, _ = setup_dummy_embedder_deployment
    deployment_handle = await AanaDeploymentHandle.create(deployment_name)
    component = AanaDeploymentComponent(deployment_handle, "embed")
    result = component.run(texts=["Hello, world!", "Roses are red", "Violets are blue"])

    assert "embeddings" in result
    assert isinstance(result["embeddings"], np.ndarray)
    assert result["embeddings"].shape == (3, 8)


@pytest.mark.asyncio
async def test_haystack_pipeline(setup_dummy_embedder_deployment):
    """Tests haystack wrapper in a pipeline."""
    deployment_name, _ = setup_dummy_embedder_deployment

    @haystack.component
    class DocumentCombinerComponent:
        @haystack.component.output_types(documents=list[Document])
        def run(self, texts: list[str], embeddings: np.ndarray):
            documents = [
                Document(content=text, embedding=embedding)
                for text, embedding in zip(texts, embeddings, strict=True)
            ]
            return {"documents": documents}

    embedder_handle = await AanaDeploymentHandle.create(deployment_name)
    embedder = AanaDeploymentComponent(embedder_handle, "embed")

    text_cleaner = TextCleaner(
        convert_to_lowercase=False, remove_punctuation=True, remove_numbers=True
    )

    document_combiner = DocumentCombinerComponent()

    pipeline = haystack.Pipeline()
    pipeline.add_component("text_cleaner", text_cleaner)
    pipeline.add_component("dummy_embedder", embedder)
    pipeline.add_component("document_combiner", document_combiner)
    pipeline.connect("text_cleaner.texts", "dummy_embedder.texts")
    pipeline.connect("dummy_embedder.embeddings", "document_combiner.embeddings")
    pipeline.connect("text_cleaner.texts", "document_combiner.texts")

    texts = ["Hello, world!", "Roses are red", "Violets are blue"]

    output = pipeline.run({"text_cleaner": {"texts": texts}})
    # {'document_combiner':
    #   {'documents':
    #       [Document(id=489009ee9ce0fd7f2aa38658ff7885fe8c2ea70fab5a711b14ee9d5eb65b2843, content: 'Hello world', embedding: vector of size 8),
    #        Document(id=489009ee9ce0fd7f2aa38658ff7885fe8c2ea70fab5a711b14ee9d5eb65b2843, content: 'Roses red', embedding: vector of size 8),
    #        Document(id=489009ee9ce0fd7f2aa38658ff7885fe8c2ea70fab5a711b14ee9d5eb65b2843, content: 'Violets blue', embedding: vector of size 8)
    #       ]
    #   }
    # }

    assert "document_combiner" in output
    assert "documents" in output["document_combiner"]
    assert len(output["document_combiner"]["documents"]) == 3
    assert all(
        isinstance(doc, Document) for doc in output["document_combiner"]["documents"]
    )


@pytest.mark.asyncio
async def test_haystack_wrapper_fails(setup_dummy_embedder_deployment):
    """Tests that haystack wrapper raises if method_name is missing."""
    deployment_name, _ = setup_dummy_embedder_deployment
    missing_method_name = "does_not_exist"
    deployment_handle = await AanaDeploymentHandle.create(deployment_name)
    with pytest.raises(AttributeError):
        AanaDeploymentComponent(deployment_handle, missing_method_name)

    failing_method_name = "embed_fail"
    component = AanaDeploymentComponent(deployment_handle, failing_method_name)
    with pytest.raises(Exception):  # noqa: B017
        component.run(texts=["Hello, world!", "Roses are red", "Violets are blue"])



import urllib.request
from contextlib import ExitStack
from pathlib import Path

import requests
from tqdm import tqdm

from aana.configs.settings import settings
from aana.exceptions.io import DownloadException
from aana.utils.file import get_sha256_hash_file

__all__ = ["download_model", "download_file"]


# Issue-Enable HF download: https://github.com/mobiusml/aana_sdk/issues/65
# model download from a url and checking SHA sum of the URL.
def download_model(
    url: str, model_hash: str = "", model_path: Path | None = None, check_sum=True
) -> Path:
    """Download a model from a URL.

    Args:
        url (str): the URL of the file to download
        model_hash (str): hash of the model file for checking sha256 hash if checksum is True
        model_path (Path): optional model path where it needs to be downloaded
        check_sum (bool): boolean to mention whether to check SHA-256 sum or not

    Returns:
        Path: the downloaded file path

    Raises:
        DownloadException: Request does not succeed.
    """
    if model_path is None:
        model_dir = settings.model_dir
        if not model_dir.exists():
            model_dir.mkdir(parents=True)
        model_path = model_dir / "model.bin"

    if model_path.exists() and not model_path.is_file():
        raise RuntimeError(f"Not a regular file: {model_path}")  # noqa: TRY003

    if not model_path.exists():
        try:
            with ExitStack() as stack:
                source = stack.enter_context(urllib.request.urlopen(url))  # noqa: S310
                output = stack.enter_context(Path.open(model_path, "wb"))

                loop = tqdm(
                    total=int(source.info().get("Content-Length")),
                    ncols=80,
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                )

                with loop:
                    while True:
                        buffer = source.read(8192)
                        if not buffer:
                            break

                        output.write(buffer)
                        loop.update(len(buffer))
        except Exception as e:
            raise DownloadException(url) from e

    model_sha256_hash = get_sha256_hash_file(model_path)
    if check_sum and model_sha256_hash != model_hash:
        checksum_error = "Model has been downloaded but the SHA256 checksum does not not match. Please retry loading the model."
        raise RuntimeError(f"{checksum_error}")

    return model_path


def download_file(url: str) -> bytes:
    """Download a file from a URL.

    Args:
        url (str): the URL of the file to download

    Returns:
        bytes: the file content

    Raises:
        DownloadException: Request does not succeed.
    """
    # TODO: add retries, check status code, etc.: add issue link
    try:
        response = requests.get(url)  # noqa: S113 TODO : add issue link
    except Exception as e:
        raise DownloadException(url) from e
    return response.content



# ruff: noqa: S101
from importlib import resources

import pytest

from aana.core.models.chat import ChatMessage
from aana.core.models.image import Image
from aana.core.models.image_chat import ImageChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.idefics_2_deployment import Idefics2Config, Idefics2Deployment
from aana.tests.utils import verify_deployment_results
from aana.utils.core import get_object_hash

deployments = [
    (
        "idefics_2_8b_deployment",
        Idefics2Deployment.options(
            num_replicas=1,
            ray_actor_options={"num_gpus": 0.85},
            user_config=Idefics2Config(
                model="HuggingFaceM4/idefics2-8b",
                dtype=Dtype.FLOAT16,
                default_sampling_params=SamplingParams(
                    temperature=1.0, max_tokens=256, kwargs={"diversity_penalty": 0.0}
                ),
            ).model_dump(mode="json"),
        ),
    )
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestIdefics2Deployment:
    """Test Idefics 2 deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "prompt, image_name",
        [("Who is the painter of the image?", "Starry_Night.jpeg")],
    )
    async def test_idefics2_deployment_chat(self, setup_deployment, prompt, image_name):
        """Test Idefics 2 deployments."""
        deployment_name, handle_name, app = setup_deployment
        handle = await AanaDeploymentHandle.create(handle_name)

        prompt_hash = get_object_hash(prompt)
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "idefics"
            / f"{deployment_name}_{image_name}_{prompt_hash}.json"
        )

        image = Image(path=resources.files("aana.tests.files.images") / image_name)
        dialog = ImageChatDialog.from_prompt(prompt=prompt, images=[image])

        # test chat method
        output = await handle.chat(dialog=dialog)
        output_message = output["message"]

        assert isinstance(output_message, ChatMessage)
        assert output_message.role == "assistant"
        verify_deployment_results(expected_output_path, output_message.content)

        # test chat_batch method
        batch_size = 2
        dialogs = [dialog] * batch_size
        outputs = await handle.chat_batch(dialogs=dialogs)
        assert len(outputs) == batch_size
        for output in outputs:
            output_message = output["message"]

            assert isinstance(output_message, ChatMessage)
            assert output_message.role == "assistant"
            verify_deployment_results(expected_output_path, output_message.content)

        # test chat_stream method
        stream = handle.chat_stream(dialog=dialog)
        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)



# ruff: noqa: S101
import pytest
from jinja2 import TemplateError
from transformers import AutoTokenizer

from aana.core.chat.chat_template import apply_chat_template
from aana.core.models.chat import ChatDialog, ChatMessage

dialog = ChatDialog(
    messages=[
        ChatMessage(
            role="system",
            content="You are a friendly chatbot who always responds in the style of a pirate",
        ),
        ChatMessage(
            role="user",
            content="How many helicopters can a human eat in one sitting?",
        ),
        ChatMessage(role="assistant", content="I don't know, how many?"),
        ChatMessage(role="user", content="One, but only if they're really hungry!"),
    ]
)


def test_chat_template_blank():
    """Test that applying a blank chat template to a dialog fails."""
    checkpoint = "NousResearch/Llama-2-7b-hf"  # Doesn't have a chat template
    tokenizer = AutoTokenizer.from_pretrained(checkpoint)

    with pytest.raises(ValueError):
        apply_chat_template(tokenizer, dialog)


def test_chat_template_custom():
    """Test that applying a custom chat template to a dialog succeeds."""
    checkpoint = "NousResearch/Llama-2-7b-hf"  # Doesn't have a chat template
    tokenizer = AutoTokenizer.from_pretrained(checkpoint)

    prompt = apply_chat_template(
        tokenizer, dialog, "llama2"
    )  # Apply custom chat template "llama2"
    assert prompt == (
        "<s>[INST] <<SYS>>\\nYou are a friendly chatbot who always responds in the style "
        "of a pirate\\n<</SYS>>\\n\\nHow many helicopters can a human eat in one sitting? "
        "[/INST] I don't know, how many? </s><s>[INST] One, but only if they're really hungry! [/INST]"
    )


def test_chat_messages_without_system():
    """Test that applying a custom chat template to a dialog without a system message succeeds."""
    dialog = ChatDialog(
        messages=[
            ChatMessage(
                role="user",
                content="How many helicopters can a human eat in one sitting?",
            ),
            ChatMessage(role="assistant", content="I don't know, how many?"),
            ChatMessage(role="user", content="One, but only if they're really hungry!"),
        ]
    )

    checkpoint = "NousResearch/Llama-2-7b-hf"  # Doesn't have a chat template
    tokenizer = AutoTokenizer.from_pretrained(checkpoint)

    prompt = apply_chat_template(
        tokenizer, dialog, "llama2"
    )  # Apply custom chat template "llama2"

    assert prompt == (
        "<s>[INST] How many helicopters can a human eat in one sitting? [/INST] I don't know, how many? </s>"
        "<s>[INST] One, but only if they're really hungry! [/INST]"
    )


def test_chat_messages_wrong_role():
    """Test that applying a custom chat template to a dialog with assistant message first fails."""
    dialog = ChatDialog(
        messages=[
            ChatMessage(
                role="assistant",
                content="How many helicopters can a human eat in one sitting?",
            ),
            ChatMessage(role="user", content="I don't know, how many?"),
            ChatMessage(
                role="assistant", content="One, but only if they're really hungry!"
            ),
        ]
    )

    checkpoint = "NousResearch/Llama-2-7b-hf"  # Doesn't have a chat template
    tokenizer = AutoTokenizer.from_pretrained(checkpoint)

    with pytest.raises(TemplateError):
        apply_chat_template(
            tokenizer, dialog, "llama2"
        )  # Apply custom chat template "llama2"


def test_from_list():
    """Test that the ChatDialog.from_list method works as expected."""
    messages = [
        {
            "role": "user",
            "content": "Can you provide ways to eat combinations of bananas and dragonfruits?",
        },
        {
            "role": "assistant",
            "content": (
                "Sure! Here are some ways to eat bananas and dragonfruits together: "
                "1. Banana and dragonfruit smoothie: Blend bananas and dragonfruits together with some milk and honey. "
                "2. Banana and dragonfruit salad: Mix sliced bananas and dragonfruits together with some lemon juice and honey."
            ),
        },
        {"role": "user", "content": "What about solving an 2x + 3 = 7 equation?"},
    ]
    dialog = ChatDialog.from_list(messages)
    assert len(dialog.messages) == len(messages)
    for message, expected in zip(dialog.messages, messages, strict=False):
        assert message.role == expected["role"]
        assert message.content == expected["content"]



from enum import Enum
from typing import Any

import torch
from hqq.core.quantize import BaseQuantizeConfig
from hqq.core.quantize import HQQBackend as HQQBackendKernel
from hqq.models.hf.base import AutoHQQHFModel
from hqq.utils.generation_hf import patch_model_for_compiled_runtime
from hqq.utils.patching import (
    HQQLinear,
    patch_add_quant_config,
    patch_linearlayers,
    prepare_for_inference,
)
from pydantic import BaseModel, ConfigDict, Field
from ray import serve
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
)

from aana.core.models.base import pydantic_protected_fields
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.hf_pipeline_deployment import CustomConfig
from aana.deployments.hf_text_generation_deployment import (
    BaseHfTextGenerationDeployment,
)


class HqqBackend(str, Enum):
    """HQQ Backend types.

    Possible values are "torchao_int4" and "bitblas".

    Attributes:
        TORCHAO_INT4 (str): torchao_int4
        BITBLAS (str): bitblas
    """

    TORCHAO_INT4 = "torchao_int4"
    BITBLAS = "bitblas"


class HqqTexGenerationConfig(BaseModel):
    """The configuration for the HQQ text generation deployment.

    Attributes:
        model_id (str): The model ID on Hugging Face.
        quantize_on_fly: Whether to quantize the model or it is already pre-quantized. Defaults to False.
        backend (HqqBackend): The backend library to use. Defaults to HqqBackend.BITBLAS.
        compile (bool): Whether to compile the model with torch.compile. Defaults to False.
        dtype (Dtype): The data type. Defaults to Dtype.AUTO.
        quantization_config (dict): The quantization configuration.
        model_kwargs (CustomConfig): The extra model keyword arguments. Defaults to {}.
        default_sampling_params (SamplingParams): The default sampling parameters.
            Defaults to SamplingParams(temperature=0, max_tokens=256).
        chat_template (str | None): The name of the chat template. If not provided, the chat template
            from the model will be used. Some models may not have a chat template. Defaults to None.
    """

    model_id: str
    quantize_on_fly: bool = False
    backend: HqqBackend = HqqBackend.BITBLAS
    compile: bool = True
    dtype: Dtype = Field(default=Dtype.AUTO)
    quantization_config: CustomConfig = BaseQuantizeConfig()
    model_kwargs: CustomConfig = {}

    default_sampling_params: SamplingParams = SamplingParams(
        temperature=0, max_tokens=512
    )
    chat_template: str | None = None

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment
class HqqTextGenerationDeployment(BaseHfTextGenerationDeployment):
    """Deployment to serve Hugging Face text generation models with Half-Quadratic Quantization (HQQ)."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and tokenizer from HuggingFace and if needed quantizes the model using HQQ on the fly.

        The configuration should conform to the HqqTexGenerationConfig schema.
        """
        config_obj = HqqTexGenerationConfig(**config)
        self.model_id = config_obj.model_id
        self.backend = config_obj.backend
        self.quantization_config = config_obj.quantization_config
        self.dtype = config_obj.dtype
        self.model_kwargs = config_obj.model_kwargs
        self.default_sampling_params = config_obj.default_sampling_params

        default_device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.model_kwargs.get("device_map", default_device)

        if self.backend == HqqBackend.BITBLAS:
            try:
                import bitblas  # noqa: F401
            except ImportError as e:
                raise ImportError(  # noqa: TRY003
                    "Failed to import the BitBLAS but HQQ is configured to use BitBLAS backend. "
                    "Check if BitBlas is correctly installed "
                    "if you want to use the bitblas backend (https://github.com/microsoft/BitBLAS)."
                ) from e

        if self.dtype == Dtype.AUTO:
            if self.backend == HqqBackend.TORCHAO_INT4:
                self.dtype = Dtype.BFLOAT16
            elif self.backend == HqqBackend.BITBLAS:
                self.dtype = Dtype.FLOAT16
            else:
                self.dtype = Dtype.BFLOAT16

        if config_obj.quantize_on_fly:
            self.model_kwargs.pop("device_map", None)
            self.model = AutoModelForCausalLM.from_pretrained(
                self.model_id, torch_dtype=self.dtype.to_torch(), **self.model_kwargs
            )
            AutoHQQHFModel.quantize_model(
                self.model,
                quant_config=self.quantization_config,
                device=self.device,
                compute_dtype=self.dtype.to_torch(),
            )
        else:
            self.model_kwargs["device"] = self.device
            self.model = AutoHQQHFModel.from_quantized(
                self.model_id, compute_dtype=self.dtype.to_torch(), **self.model_kwargs
            )
            patch_linearlayers(
                self.model, patch_add_quant_config, self.quantization_config
            )

        HQQLinear.set_backend(HQQBackendKernel.PYTORCH)
        self.model.generation_config.cache_implementation = "static"
        prepare_for_inference(self.model, backend=self.backend)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
        self.chat_template_name = config_obj.chat_template
        if config_obj.compile:
            patch_model_for_compiled_runtime(self.model, self.tokenizer, warmup=True)



from typing import Annotated, Any, Literal

from pydantic import BaseModel, ConfigDict, Field

from aana.core.models.image import Image

Role = Literal["system", "user", "assistant"]


class TextContent(BaseModel):
    """Text content for a chat message.

    Attributes:
        type (Literal["text"]): the type of the content, always "text"
        text (str): the text of the message
    """

    type: Literal["text"] = "text"
    text: str


class ImageContent(BaseModel):
    """Image content for a chat message.

    Attributes:
        type (Literal["image"]): the type of the content, always "image"
        image (Image): the image
    """

    type: Literal["image"] = "image"
    image: Image

    model_config = ConfigDict(arbitrary_types_allowed=True)


Content = Annotated[
    TextContent | ImageContent,
    Field(description="The content of the message, either text or image."),
]


class ImageChatMessage(BaseModel):
    """A chat message with image support.

    Attributes:
        content (list[Content]): the content of the message
        role (Role): the role of the message
    """

    content: list[Content]
    role: Role
    model_config = ConfigDict(
        json_schema_extra={"description": "A chat message with image support."}
    )


class ImageChatDialog(BaseModel):
    """A chat dialog with image support.

    Attributes:
        messages (list[ImageChatMessage]): the list of messages
    """

    messages: list[ImageChatMessage]
    model_config = ConfigDict(
        json_schema_extra={"description": "A chat dialog with image support."}
    )

    @classmethod
    def from_list(cls, messages: list[dict[str, Any]]) -> "ImageChatDialog":
        """Create an ImageChatDialog from a list of messages.

        Args:
            messages (list[dict[str, str]]): the list of messages

        Returns:
            ImageChatDialog: the chat dialog
        
        Example:
        ```
        messages = [
            {
                "content": [
                    { "type": "image", "image": Image(...) },
                    { "type": "text", "text": "..." }
                ],
                "role": "system"
            },
            {
                "content": [
                    { "type": "image", "image": Image(...) },
                    { "type": "text", "text": "..." }
                ],
                "role": "user"
            }
        ]
        dialog = ImageChatDialog.from_list(messages)
        ```
        """
        return ImageChatDialog(
            messages=[ImageChatMessage(**message) for message in messages]
        )

    @classmethod
    def from_prompt(cls, prompt: str, images: list[Image]) -> "ImageChatDialog":
        """Create an ImageChatDialog from a prompt and a list of images.

        Args:
            prompt (str): the prompt
            images (list[Image]): the list of images

        Returns:
            ImageChatDialog: the chat dialog
        """
        content: list[Content] = [ImageContent(image=image) for image in images]
        content.append(TextContent(text=prompt))

        return ImageChatDialog(
            messages=[ImageChatMessage(content=content, role="user")]
        )

    def to_objects(self) -> tuple[list[dict], list[Image]]:
        """Convert ImageChatDialog to messages and images.

        Returns:
            tuple[list[dict], list[Image]]: the messages and the images
        """
        dialog_dict = self.model_dump(
            exclude={"messages": {"__all__": {"content": {"__all__": {"image"}}}}}
        )
        messages = dialog_dict["messages"]
        # images = []
        # for message in self.messages:
        #     for content in message.content:
        #         if content.type == "image":
        #             images.append(content.image)
        images = [content.image for message in self.messages for content in message.content if content.type == "image"]

        return messages, images



import asyncio
from queue import Empty


async def async_streamer_adapter(streamer):
    """Adapt the TextIteratorStreamer to an async generator."""
    while True:
        try:
            for item in streamer:
                yield item
            break
        except Empty:
            # wait for the next item
            await asyncio.sleep(0.01)



# ruff: noqa: S101, S113
import json
from collections.abc import AsyncGenerator
from typing import TypedDict

import requests
from ray import serve

from aana.api.api_generation import Endpoint
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment


@serve.deployment
class Lowercase(BaseDeployment):
    """Ray Serve deployment that returns the lowercase version of a text."""

    async def lower_stream(self, text: str) -> AsyncGenerator[dict, None]:
        """Lowercase the text and yield the lowercase text in chunks of 4 characters.

        Args:
            text (str): The text to lowercase

        Yields:
            dict: The lowercase text in chunks of 4 characters
        """
        chunk_size = 4
        for i in range(0, len(text), chunk_size):
            t = text[i : i + chunk_size]
            yield {"text": t.lower()}


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: str


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(self, text: str) -> AsyncGenerator[LowercaseEndpointOutput, None]:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        async for chunk in self.lowercase_handle.lower_stream(text=text):
            yield {"text": chunk["text"]}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    },
]

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    }
]


def test_app_streaming(create_app):
    """Test the Ray Serve app with streaming enabled."""
    aana_app = create_app(deployments=deployments, endpoints=endpoints)

    port = aana_app.port
    route_prefix = ""

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}{route_prefix}/api/ready")
    assert response.status_code == 200
    assert response.json() == {"ready": True}

    # Test lowercase endpoint
    text = "Hello World, this is a test."
    data = {"text": text}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
        stream=True,
    )
    assert response.status_code == 200

    lowercase_text = ""
    offset = 0
    for chunk in response.iter_content(chunk_size=None):
        json_data = json.loads(chunk)
        lowercase_text_chunk = json_data["text"]
        lowercase_text += lowercase_text_chunk

        chunk_size = len(lowercase_text_chunk)
        expected_chunk = text[offset : offset + chunk_size]

        assert lowercase_text_chunk == expected_chunk.lower()

        offset += chunk_size

    assert lowercase_text == text.lower()



from aana.storage.repository.base import BaseRepository
from aana.storage.repository.caption import CaptionRepository
from aana.storage.repository.media import MediaRepository
from aana.storage.repository.transcript import TranscriptRepository
from aana.storage.repository.video import VideoRepository

__all__ = [
    "BaseRepository",
    "MediaRepository",
    "VideoRepository",
    "CaptionRepository",
    "TranscriptRepository",
]



from collections.abc import AsyncGenerator
from threading import Thread
from typing import Any

import torch
import transformers
from pydantic import BaseModel, ConfigDict
from ray import serve
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    TextIteratorStreamer,
)

from aana.core.models.base import merged_options, pydantic_protected_fields
from aana.core.models.sampling import SamplingParams
from aana.deployments.base_deployment import exception_handler
from aana.deployments.base_text_generation_deployment import (
    BaseTextGenerationDeployment,
    LLMOutput,
)
from aana.deployments.hf_pipeline_deployment import CustomConfig
from aana.exceptions.runtime import InferenceException, PromptTooLongException
from aana.utils.streamer import async_streamer_adapter


class HfTextGenerationConfig(BaseModel):
    """The configuration for the Hugging Face text generation deployment.

    Attributes:
        model_id (str): The model ID on Hugging Face.
        model_kwargs (CustomConfig): The extra model keyword arguments. Defaults to {}.
        default_sampling_params (SamplingParams): The default sampling parameters.
            Defaults to SamplingParams(temperature=0, max_tokens=256).
        chat_template (str | None): The name of the chat template. If not provided, the chat template
            from the model will be used. Some models may not have a chat template. Defaults to None.
    """

    model_id: str
    model_kwargs: CustomConfig = {}
    default_sampling_params: SamplingParams = SamplingParams(
        temperature=0, max_tokens=256
    )
    chat_template: str | None = None

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class BaseHfTextGenerationDeployment(BaseTextGenerationDeployment):
    """Base class for Hugging Face text generation deployments."""

    @exception_handler
    async def generate_stream(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate text stream.

        Args:
            prompt (str): the prompt
            sampling_params (SamplingParams | None): the sampling parameters

        Yields:
            LLMOutput: the generated text
        """
        # Set the seed to make the results reproducible
        transformers.set_seed(42)

        prompt = str(prompt)

        if sampling_params is None:
            sampling_params = self.default_sampling_params
        else:
            sampling_params = merged_options(
                self.default_sampling_params, sampling_params
            )

        prompt_input = self.tokenizer(
            prompt, return_tensors="pt", add_special_tokens=False
        )

        if prompt_input["input_ids"].shape[-1] > self.tokenizer.model_max_length:
            raise PromptTooLongException(
                prompt_len=prompt_input["input_ids"].shape[-1],
                max_len=self.tokenizer.model_max_length,
            )

        try:
            with torch.no_grad():
                prompt_input.to(self.model.device)
                streamer = TextIteratorStreamer(
                    self.tokenizer,
                    timeout=0,
                    skip_prompt=True,
                    skip_special_tokens=True,
                )
                generation_kwargs = dict(
                    prompt_input,
                    streamer=streamer,
                    max_new_tokens=sampling_params.max_tokens,
                    top_k=sampling_params.top_k,
                    top_p=sampling_params.top_p,
                    temperature=sampling_params.temperature,
                    num_return_sequences=1,
                    eos_token_id=self.tokenizer.eos_token_id,
                    repetition_penalty=sampling_params.repetition_penalty,
                    **sampling_params.kwargs,
                )
                if sampling_params.temperature == 0:
                    generation_kwargs["do_sample"] = False
                    generation_kwargs["temperature"] = None
                generation_thread = Thread(
                    target=self.model.generate, kwargs=generation_kwargs
                )
                generation_thread.start()

                async_streamer = async_streamer_adapter(streamer)
                async for new_text in async_streamer:
                    yield LLMOutput(text=new_text)

                # clean up
                prompt_input.to("cpu")
                del streamer
        except Exception as e:
            raise InferenceException(model_name=self.model_id) from e


@serve.deployment
class HfTextGenerationDeployment(BaseHfTextGenerationDeployment):
    """Deployment to serve Hugging Face text generation models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and tokenizer from HuggingFace.

        The configuration should conform to the HfTextGenerationConfig schema.
        """
        config_obj = HfTextGenerationConfig(**config)
        self.model_id = config_obj.model_id
        self.model_kwargs = config_obj.model_kwargs
        self.default_sampling_params = config_obj.default_sampling_params

        self.model_kwargs["device_map"] = self.model_kwargs.get("device_map", "auto")
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_id, **self.model_kwargs
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
        self.chat_template_name = config_obj.chat_template



# ruff: noqa: S101

import random
import uuid

import pytest

from aana.core.models.captions import Caption
from aana.exceptions.db import NotFoundException
from aana.storage.repository.caption import CaptionRepository


@pytest.fixture(scope="function")
def dummy_caption():
    """Creates a dummy caption for testing."""
    caption = Caption(f"This is a caption {uuid.uuid4()}")
    frame_id = random.randint(0, 100)  # noqa: S311
    timestamp = random.random()  # noqa: S311
    return caption, frame_id, timestamp


def test_save_caption(db_session, dummy_caption):
    """Tests saving a caption."""
    caption, frame_id, timestamp = dummy_caption
    model_name = "blip2"

    caption_repo = CaptionRepository(db_session)
    caption_entity = caption_repo.save(
        model_name=model_name,
        caption=caption,
        frame_id=frame_id,
        timestamp=timestamp,
    )
    caption_id = caption_entity.id

    caption_entity = caption_repo.read(caption_id)
    assert caption_entity.model == model_name
    assert caption_entity.frame_id == frame_id
    assert caption_entity.timestamp == timestamp
    assert caption_entity.caption == caption

    caption_repo.delete(caption_id)
    with pytest.raises(NotFoundException):
        caption_repo.read(caption_id)


def test_save_all_captions(db_session, dummy_caption):
    """Tests saving all captions."""
    captions, frame_ids, timestamps = [], [], []
    for _ in range(3):
        caption, frame_id, timestamp = dummy_caption
        captions.append(caption)
        frame_ids.append(frame_id)
        timestamps.append(timestamp)
    model_name = "blip2"

    caption_repo = CaptionRepository(db_session)
    caption_entities = caption_repo.save_all(
        model_name=model_name,
        captions=captions,
        timestamps=timestamps,
        frame_ids=frame_ids,
    )
    assert len(caption_entities) == len(captions)

    caption_ids = [caption_entity.id for caption_entity in caption_entities]
    for caption_id, caption, frame_id, timestamp in zip(
        caption_ids, captions, frame_ids, timestamps, strict=True
    ):
        caption_entity = caption_repo.read(caption_id)

        assert caption_entity.model == model_name
        assert caption_entity.frame_id == frame_id
        assert caption_entity.timestamp == timestamp
        assert caption_entity.caption == caption

    # delete all captions
    for caption_id in caption_ids:
        caption_repo.delete(caption_id)
        with pytest.raises(NotFoundException):
            caption_repo.read(caption_id)



from pydantic import BaseModel, ConfigDict, Field

__all__ = ["TimeInterval"]


class TimeInterval(BaseModel):
    """Pydantic schema for TimeInterval.

    Attributes:
        start (float): Start time in seconds
        end (float): End time in seconds
    """

    start: float = Field(ge=0.0, description="Start time in seconds")
    end: float = Field(ge=0.0, description="End time in seconds")
    model_config = ConfigDict(
        json_schema_extra={
            "description": "Time interval in seconds",
        }
    )



from collections.abc import Callable
from types import CoroutineType, NoneType
from typing import get_type_hints

from haystack import component

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.utils.asyncio import run_async
from aana.utils.typing import is_typed_dict


def typehints_to_component_types(
    typehints: dict[str, type],
) -> tuple[dict[str, type], dict[str, type]]:
    """Converts a type hint dictionary into something that can be consumed by Haystack.

    Arguments:
        typehints (dict): a dict returned by `typing.get_type_hints()`

    Returns:
        tuple[dict[str,type], dict[str,type]]: Suitable for passing to the haystack component interface.
    """
    output_types = typehints_to_output_types(typehints.pop("return", NoneType))
    input_types = typehints_to_input_types(typehints)
    return input_types, output_types


def typehints_to_output_types(typehint: type) -> dict[str, type]:
    """Converts a return type hint into something that can be consumed by Haystack.

    Arguments:
        typehint: Return type from `typing.get_type_hints()["return"]`

    Returns:
        dict[str, type]: Something that can be passed to `haystack.component.set_output_types()`
    """
    # If no return value, return an empty dictionary
    if typehint is None:
        return {}
    # If annotation is a TypedDict, turn it into a regular dict
    if is_typed_dict(typehint):
        return get_type_hints(typehint)
    # If it's a class with annotations, return the __annotations__ dict
    annotations = getattr(typehint, "__annotations__", None)
    if annotations:
        return annotations
    # Otherwise wrap it into a dictionary
    return {"return": typehint}


def typehints_to_input_types(typehints: dict[str, type]) -> dict[str, type]:
    """Converts a typehint dictionary into something that can be used by Haystack.

    Arguments:
        typehints (dict[str, type]): Type hints from `typing.get_type_hints()` *minus the "return" key*

    Returns:
        dict[str, type]: Something that can be consumed by `haystack.set_input_types()`
    """
    # If typehint is None, or an empty dict, return an empty dict
    if not typehints:
        return {}
    # Otherwise just return the input
    return typehints


@component
class AanaDeploymentComponent:
    """Wrapper for Aana deployments to run as HayStack Components.

    Example:
        ```python
        deployment_handle = await AanaDeploymentHandle.create("my_deployment")
        haystack_component = AanaDeploymentComponent(deployment_handle, "my_method")
        haystack_component.warm_up()  # This is currently a no-op, but subject to change.
        component_result = haystack_component.run(my_input_prompt="This is an input prompt")
        ```
    """

    _deployment_handle: AanaDeploymentHandle
    _run_method: Callable
    _warm: bool

    def __init__(self, deployment_handle: AanaDeploymentHandle, method_name: str):
        """Constructor.

        Arguments:
            deployment_handle (AanaDeploymentHandle): the Aana Ray deployment to be wrapped (must be a class Deployment)
            method_name (str): the name of the method on the deployment to call inside the component's `run()` method.
        """
        self._deployment_handle = deployment_handle

        # Determine input and output types for `run()`
        # Will raise if the function is not defined (e.g. if you pass a function deployment)
        self.run_method = self._get_method(method_name)
        if not self.run_method:
            raise AttributeError(name=method_name, obj=self._deployment_handle)
        hints = get_type_hints(self.run_method)
        input_types, output_types = typehints_to_component_types(hints)
        # The functions `set_input_types()` and `set_output_types()`
        # take an positional instance argument and keyword arguments
        component.set_input_types(self, **input_types)
        component.set_output_types(self, **output_types)

    def warm_up(self):
        """Warms up the deployment to a ready state.

        As we run off an existing deployment handle, this is currently a no-op.
        """
        self._warm = True

    def run(self, *args, **kwargs):
        """Run the component. This is the primary interface for Haystack Components.

        Arguments:
            *args: the arguments to pass to the deployment run function
            **kwargs: the keyword arguments to pass to the deployment run function

        Returns:
            The return value of the deployment's run function
        """
        # Function may (must?) be a coroutine. Resolve it if so.
        return run_async(self._call(*args, **kwargs))

    def _call(self, *args, **kwargs) -> CoroutineType:
        """Calls the deployment's run method. Not public, use the `run()` method."""
        return self.run_method(*args, **kwargs)  # type: ignore

    def _get_method(self, method_name: str) -> Callable | None:
        """Gets a handle to the method specified by the constructor."""
        instance_method_handle: Callable | None = getattr(
            self._deployment_handle, method_name, None
        )
        return instance_method_handle



# ruff: noqa: S101, S113
from typing import Any

import pytest
from ray import serve

from aana.deployments.base_deployment import BaseDeployment
from aana.exceptions.runtime import FailedDeployment, InsufficientResources


@serve.deployment
class DummyFailingDeployment(BaseDeployment):
    """Simple deployment that fails on initialization."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration to the deployment and initialize it."""
        raise Exception("Dummy exception")  # noqa: TRY002, TRY003


@serve.deployment
class Lowercase(BaseDeployment):
    """Simple deployment that lowercases the text."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration to the deployment and initialize it."""
        pass

    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        return {"text": [t.lower() for t in text]}


def test_failed_deployment(create_app):
    """Test that a failed deployment raises a FailedDeployment exception."""
    deployments = [
        {
            "name": "deployment",
            "instance": DummyFailingDeployment.options(num_replicas=1, user_config={}),
        }
    ]
    with pytest.raises(FailedDeployment):
        create_app(deployments, [])


def test_insufficient_resources(create_app):
    """Test that deployment fails when there are insufficient resources to deploy."""
    deployments = [
        {
            "name": "deployment",
            "instance": Lowercase.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 100},  # requires 100 GPUs
                user_config={},
            ),
        }
    ]
    with pytest.raises(InsufficientResources):
        create_app(deployments, [])



import uuid
from enum import Enum

from sqlalchemy import (
    JSON,
    UUID,
    PickleType,
)
from sqlalchemy.orm import Mapped, mapped_column

from aana.storage.models.base import BaseEntity, TimeStampEntity, timestamp


class Status(str, Enum):
    """Enum for task status."""

    CREATED = "created"
    ASSIGNED = "assigned"
    COMPLETED = "completed"
    RUNNING = "running"
    FAILED = "failed"
    NOT_FINISHED = "not_finished"


class TaskEntity(BaseEntity, TimeStampEntity):
    """Table for task items."""

    __tablename__ = "tasks"

    id: Mapped[uuid.UUID] = mapped_column(
        UUID, primary_key=True, default=uuid.uuid4, comment="Task ID"
    )
    endpoint: Mapped[str] = mapped_column(
        nullable=False, comment="The endpoint to which the task is assigned"
    )
    data = mapped_column(PickleType, nullable=False, comment="Data for the task")
    status: Mapped[Status] = mapped_column(
        default=Status.CREATED,
        comment="Status of the task",
    )
    priority: Mapped[int] = mapped_column(
        nullable=False, default=0, comment="Priority of the task (0 is the lowest)"
    )
    assigned_at: Mapped[timestamp | None] = mapped_column(
        comment="Timestamp when the task was assigned",
    )
    completed_at: Mapped[timestamp | None] = mapped_column(
        server_default=None,
        comment="Timestamp when the task was completed",
    )
    progress: Mapped[float] = mapped_column(
        nullable=False, default=0.0, comment="Progress of the task in percentage"
    )
    result: Mapped[dict | None] = mapped_column(
        JSON, comment="Result of the task in JSON format"
    )
    num_retries: Mapped[int] = mapped_column(
        nullable=False, default=0, comment="Number of retries"
    )

    def __repr__(self) -> str:
        """String representation of the task."""
        return (
            f"<TaskEntity(id={self.id}, "
            f"endpoint={self.endpoint}, status={self.status}, "
            f"priority={self.priority}, progress={self.progress}, "
            f"num_retries={self.num_retries}, "
            f"updated_at={self.updated_at})>"
        )



import traceback

from fastapi import Request
from pydantic import ValidationError
from ray.exceptions import RayTaskError

from aana.api.responses import AanaJSONResponse
from aana.core.models.exception import ExceptionResponseModel
from aana.exceptions.core import BaseException


async def validation_exception_handler(request: Request, exc: ValidationError):
    """This handler is used to handle pydantic validation errors.

    Args:
        request (Request): The request object
        exc (ValidationError): The validation error

    Returns:
        JSONResponse: JSON response with the error details
    """
    return AanaJSONResponse(
        status_code=422,
        content=ExceptionResponseModel(
            error="ValidationError",
            message="Validation error",
            data=exc.errors(),
        ).model_dump(),
    )


def custom_exception_handler(request: Request | None, exc_raw: Exception):
    """This handler is used to handle custom exceptions raised in the application.

    BaseException is the base exception for all the exceptions
    from the Aana application.
    Sometimes custom exception are wrapped into RayTaskError so we need to handle that as well.

    Args:
        request (Request): The request object
        exc_raw (Exception): The exception raised

    Returns:
        JSONResponse: JSON response with the error details. The response contains the following fields:
            error: The name of the exception class.
            message: The message of the exception.
            data: The additional data returned by the exception that can be used to identify the error (e.g. image path, url, model name etc.)
            stacktrace: The stacktrace of the exception.
    """
    # a BaseException can be wrapped into a RayTaskError
    if isinstance(exc_raw, RayTaskError):
        # str(e) returns whole stack trace
        # if exception is a RayTaskError
        # let's use it to get the stack trace
        stacktrace = str(exc_raw)
        # get the original exception
        exc = exc_raw.cause
    else:
        # if it is not a RayTaskError
        # then we need to get the stack trace
        stacktrace = traceback.format_exc()
        exc = exc_raw
    # get the data from the exception
    # can be used to return additional info
    # like image path, url, model name etc.
    data = exc.get_data() if isinstance(exc, BaseException) else {}
    # get the name of the class of the exception
    # can be used to identify the type of the error
    error = exc.__class__.__name__
    # get the message of the exception
    message = str(exc)
    status_code = getattr(exc, "http_status_code", 400)
    return AanaJSONResponse(
        status_code=status_code,
        content=ExceptionResponseModel(
            error=error, message=message, data=data, stacktrace=stacktrace
        ).model_dump(),
    )


async def aana_exception_handler(request: Request, exc: Exception):
    """This handler is used to handle exceptions raised by the Aana application.

    Args:
        request (Request): The request object
        exc (Exception): The exception raised

    Returns:
        JSONResponse: JSON response with the error details
    """
    return custom_exception_handler(request, exc)



import typing
from enum import Enum
from pathlib import Path

import orjson
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine

from aana.exceptions.runtime import EmptyMigrationsException
from aana.utils.core import get_module_dir
from aana.utils.json import jsonify

if typing.TYPE_CHECKING:
    from aana.configs.db import DbSettings


class DbType(str, Enum):
    """Engine types for relational database.

    Attributes:
        POSTGRESQL: PostgreSQL database.
        SQLITE: SQLite database.
    """

    POSTGRESQL = "postgresql"
    SQLITE = "sqlite"


def create_postgresql_engine(db_config: "DbSettings"):
    """Create PostgreSQL engine based on the provided configuration.

    Args:
        db_config (DbSettings): Database configuration.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    datastore_config = db_config.datastore_config
    user = datastore_config["user"]
    password = datastore_config["password"]
    host = datastore_config["host"]
    port = datastore_config["port"]
    database = datastore_config["database"]
    connection_string = (
        f"postgresql+psycopg://{user}:{password}@{host}:{port}/{database}"
    )
    return create_engine(
        connection_string,
        json_serializer=lambda obj: jsonify(obj),
        json_deserializer=orjson.loads,
        pool_size=db_config.pool_size,
        max_overflow=db_config.max_overflow,
        pool_recycle=db_config.pool_recycle,
    )


def create_sqlite_engine(db_config: "DbSettings"):
    """Create an SQLite SQLAlchemy engine based on the provided configuration.

    Args:
        db_config (DbSettings): Database configuration.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    datastore_config = db_config.datastore_config
    connection_string = f"sqlite:///{datastore_config['path']}"
    return create_engine(
        connection_string,
        json_serializer=lambda obj: jsonify(obj),
        json_deserializer=orjson.loads,
        pool_size=db_config.pool_size,
        max_overflow=db_config.max_overflow,
        pool_recycle=db_config.pool_recycle,
    )


def create_database_engine(db_config: "DbSettings"):
    """Create SQLAlchemy engine based on the provided configuration.

    Args:
        db_config (DbSettings): Database configuration.

    Returns:
        sqlalchemy.engine.Engine: SQLAlchemy engine instance.
    """
    db_type = getattr(db_config, "datastore_type", "").lower()

    if db_type == DbType.POSTGRESQL:
        return create_postgresql_engine(db_config)
    elif db_type == DbType.SQLITE:
        return create_sqlite_engine(db_config)
    else:
        raise ValueError(f"Unsupported database type: {db_type}")  # noqa: TRY003


def get_alembic_config(
    app_config, ini_file_path: Path, alembic_data_path: Path
) -> Config:
    """Produces an alembic config to run migrations programmatically."""
    engine = app_config.db_config.get_engine()
    alembic_config = Config(ini_file_path)
    alembic_config.set_main_option("script_location", str(alembic_data_path))
    config_section = alembic_config.get_section(alembic_config.config_ini_section, {})
    config_section["sqlalchemy.url"] = engine.url

    return alembic_config


def run_alembic_migrations(settings, root_path: Path | None = None):
    """Runs alembic migrations before starting up."""
    if root_path is None:
        root_path = get_module_dir("aana")

    ini_file_path = root_path / "alembic.ini"
    alembic_data_path = root_path / "alembic"
    if not alembic_data_path.exists():
        raise RuntimeError("Alembic directory does not exist.")  # noqa: TRY003
    versions_path = alembic_data_path / "versions"
    # Check if the versions directory is empty (no .py files)
    if not versions_path.exists() or not any(Path(versions_path).glob("*.py")):
        raise EmptyMigrationsException()

    alembic_config = get_alembic_config(settings, ini_file_path, alembic_data_path)
    engine = settings.db_config.get_engine()
    with engine.begin() as connection:
        alembic_config.attributes["connection"] = connection
        command.upgrade(alembic_config, "head")


def drop_all_tables(settings, root_path: Path | None = None):
    """Drops all tables in the database."""
    # TODO: only allow this in testing mode
    if root_path is None:
        root_path = get_module_dir("aana")

    ini_file_path = root_path / "alembic.ini"
    alembic_data_path = root_path / "alembic"
    if not alembic_data_path.exists():
        raise RuntimeError("Alembic directory does not exist.")  # noqa: TRY003

    alembic_config = get_alembic_config(settings, ini_file_path, alembic_data_path)
    command.downgrade(alembic_config, "base")



# ruff: noqa: S101, S113
import json
from typing import TypedDict

import requests
from haystack import Document, Pipeline
from haystack.components.preprocessors import DocumentCleaner
from haystack.components.retrievers.in_memory import InMemoryEmbeddingRetriever
from haystack.document_stores.in_memory import InMemoryDocumentStore

from aana.api.api_generation import Endpoint
from aana.deployments.haystack_component_deployment import (
    HaystackComponentDeployment,
    HaystackComponentDeploymentConfig,
    RemoteHaystackComponent,
)


class HaystackTestEndpointOutput(TypedDict):
    """The output of the Haystack test endpoint."""

    response: str


class HaystackTestEndpoint(Endpoint):
    """A test endpoint for Haystack Integration."""

    async def initialize(self):
        """Initialize the endpoint."""
        document_store = InMemoryDocumentStore(embedding_similarity_function="cosine")

        documents = [
            Document(content="My name is Wolfgang and I live in Berlin"),
            Document(content="I saw a black horse running"),
            Document(content="Germany has many big cities"),
        ]

        cleaner = DocumentCleaner()
        document_embedder = RemoteHaystackComponent("document_embedder_deployment")
        document_embedder.warm_up()

        self.indexing_pipeline = Pipeline()
        self.indexing_pipeline.add_component("cleaner", cleaner)
        self.indexing_pipeline.add_component("document_embedder", document_embedder)
        self.indexing_pipeline.connect(
            "cleaner.documents", "document_embedder.documents"
        )

        documents_with_embeddings = self.indexing_pipeline.run(
            {"cleaner": {"documents": documents}}
        )["document_embedder"]["documents"]
        document_store.write_documents(documents_with_embeddings)

        text_embedder = RemoteHaystackComponent("text_embedder_deployment")
        text_embedder.warm_up()

        self.query_pipeline = Pipeline()
        self.query_pipeline.add_component("text_embedder", text_embedder)
        self.query_pipeline.add_component(
            "retriever", InMemoryEmbeddingRetriever(document_store=document_store)
        )
        self.query_pipeline.connect(
            "text_embedder.embedding", "retriever.query_embedding"
        )
        await super().initialize()

    async def run(self, query: str) -> HaystackTestEndpointOutput:
        """Run the test endpoint for Haystack Integration."""
        result = self.query_pipeline.run({"text_embedder": {"text": query}})
        return {"response": result["retriever"]["documents"][0].content}


text_embedder_deployment = HaystackComponentDeployment.options(
    num_replicas=1,
    user_config=HaystackComponentDeploymentConfig(
        component="haystack.components.embedders.SentenceTransformersTextEmbedder",
        params={"model": "sentence-transformers/all-mpnet-base-v2"},
    ).model_dump(),
)

document_embedder_deployment = HaystackComponentDeployment.options(
    num_replicas=1,
    user_config=HaystackComponentDeploymentConfig(
        component="haystack.components.embedders.SentenceTransformersDocumentEmbedder",
        params={"model": "sentence-transformers/all-mpnet-base-v2"},
    ).model_dump(),
)


deployments = [
    {
        "name": "text_embedder_deployment",
        "instance": text_embedder_deployment,
    },
    {
        "name": "document_embedder_deployment",
        "instance": document_embedder_deployment,
    },
]

endpoints = [
    {
        "name": "haystack_test_endpoint",
        "path": "/query",
        "summary": "A test endpoint for Haystack",
        "endpoint_cls": HaystackTestEndpoint,
    }
]


def test_haystack_integration(create_app):
    """Test Haystack integration."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port

    data = {"query": "Who lives in Berlin?"}
    response = requests.post(
        f"http://localhost:{port}/query",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200
    assert response.json() == {"response": "My name is Wolfgang and I live in Berlin"}

    data = {"query": "What is the interesting fact about Germany?"}
    response = requests.post(
        f"http://localhost:{port}/query",
        data={"body": json.dumps(data)},
    )

    assert response.status_code == 200
    assert response.json() == {"response": "Germany has many big cities"}



from aana.core.models.audio import Audio
from aana.core.models.video import Video
from aana.integrations.external.av import load_audio

__all__ = ["extract_audio"]


def extract_audio(video: Video) -> Audio:
    """Extract the audio file from a Video and return an Audio object.

    Args:
        video (Video): The video file to extract audio.

    Returns:
        Audio: an Audio object containing the extracted audio.
    """
    audio_bytes = load_audio(video.path)

    # Only difference will be in path where WAV file will be stored
    # and in content but has same media_id
    return Audio(
        url=video.url,
        media_id=f"audio_{video.media_id}",
        content=audio_bytes,
        title=video.title,
        description=video.description,
    )



from typing import Any

from pydantic import BaseModel, ConfigDict


class ExceptionResponseModel(BaseModel):
    """This class is used to represent an exception response for 400 errors.

    Attributes:
        error (str): The error that occurred.
        message (str): The message of the error.
        data (Optional[Any]): The extra data that helps to debug the error.
        stacktrace (Optional[str]): The stacktrace of the error.
    """

    error: str
    message: str
    data: Any | None = None
    stacktrace: str | None = None
    model_config = ConfigDict(extra="forbid")



from collections.abc import Callable

import numpy as np
import torch
from pyannote.audio.core.io import AudioFile
from pyannote.audio.pipelines import VoiceActivityDetection
from pyannote.audio.pipelines.utils import PipelineModel
from pyannote.core import Annotation, Segment, SlidingWindowFeature


class VoiceActivitySegmentation(VoiceActivityDetection):
    """Pipeline wrapper class for performing Voice Activity Segmentation based on the detection from the VAD model."""

    def __init__(
        self,
        segmentation: PipelineModel = "pyannote/segmentation",
        device: torch.device | None = None,
        fscore: bool = False,
        use_auth_token: str | None = None,
        **inference_kwargs,
    ):
        """Initialize the pipeline with the model name and the optional device.

        Args:
            dict parameters of VoiceActivityDetection class from pyannote:
            segmentation (PipelineModel): Loaded model name.
            device (torch.device | None): Device to perform the segmentation.
            fscore (bool): Flag indicating whether to compute F-score during inference.
            use_auth_token (str | None): Optional authentication token for model access.
            inference_kwargs (dict): Optional additional arguments from VoiceActivityDetection pipeline.
        """
        super().__init__(
            segmentation=segmentation,
            device=device,
            fscore=fscore,
            use_auth_token=use_auth_token,
            **inference_kwargs,
        )

    def apply(self, file: AudioFile, hook: Callable | None = None) -> Annotation:
        """Apply voice activity detection on the audio file.

        Args:
            file (AudioFile): Processed file.
            hook (callable, optional): Hook called after each major step of the pipeline with the following signature: hook("step_name", step_artefact, file=file)

        Returns:
            segmentations (Annotation): Voice activity segmentation.
        """
        # setup hook (e.g. for debugging purposes)
        hook = self.setup_hook(file, hook=hook)

        # apply segmentation model (only if needed)
        # output shape is (num_chunks, num_frames, 1)
        if self.training:
            if self.CACHED_SEGMENTATION in file:
                segmentations = file[self.CACHED_SEGMENTATION]
            else:
                segmentations = self._segmentation(file)
                file[self.CACHED_SEGMENTATION] = segmentations
        else:
            segmentations: SlidingWindowFeature = self._segmentation(file)

        return segmentations


class BinarizeVadScores:
    """Binarize detection scores using hysteresis thresholding, with min-cut operation to ensure no segments are longer than max_duration.

    Reference:
        Gregory Gelly and Jean-Luc Gauvain. "Minimum Word Error Training of
        RNN-based Voice Activity Detection", InterSpeech 2015.

        Modified by Max Bain to include WhisperX's min-cut operation
        https://arxiv.org/abs/2303.00747

    """

    def __init__(
        self,
        onset: float = 0.5,
        offset: float | None = None,
        min_duration_on: float = 0.0,
        min_duration_off: float = 0.0,
        pad_onset: float = 0.0,
        pad_offset: float = 0.0,
        max_duration: float = float("inf"),
    ):
        """Initializes the parameters for Binarizing the VAD scores.

        Args:
            onset (float, optional):
                Onset threshold. Defaults to 0.5.
            offset (float, optional):
                Offset threshold. Defaults to `onset`.
            min_duration_on (float, optional):
                Remove active regions shorter than that many seconds. Defaults to 0s.
            min_duration_off (float, optional):
                Fill inactive regions shorter than that many seconds. Defaults to 0s.
            pad_onset (float, optional):
                Extend active regions by moving their start time by that many seconds.
                Defaults to 0s.
            pad_offset (float, optional):
                Extend active regions by moving their end time by that many seconds.
                Defaults to 0s.
            max_duration (float):
                The maximum length of an active segment, divides segment at timestamp with lowest score.
        """
        super().__init__()

        self.onset = onset
        self.offset = offset or onset

        self.pad_onset = pad_onset
        self.pad_offset = pad_offset

        self.min_duration_on = min_duration_on
        self.min_duration_off = min_duration_off

        self.max_duration = max_duration

    def __get_active_regions(self, scores: SlidingWindowFeature) -> Annotation:
        """Extract active regions from VAD scores.

        Args:
            scores (SlidingWindowFeature): Detection scores.

        Returns:
            active (Annotation): Active regions.
        """
        num_frames, num_classes = scores.data.shape
        frames = scores.sliding_window
        timestamps = [frames[i].middle for i in range(num_frames)]
        # annotation meant to store 'active' regions
        active = Annotation()
        for k, k_scores in enumerate(scores.data.T):
            label = k if scores.labels is None else scores.labels[k]

            # initial state
            start = timestamps[0]
            is_active = k_scores[0] > self.onset
            curr_scores = [k_scores[0]]
            curr_timestamps = [start]
            t = start
            for t, y in zip(timestamps[1:], k_scores[1:], strict=False):
                # currently active
                if is_active:
                    curr_duration = t - start
                    if curr_duration > self.max_duration:
                        search_after = len(curr_scores) // 2
                        # divide segment
                        min_score_div_idx = search_after + np.argmin(
                            curr_scores[search_after:]
                        )
                        min_score_t = curr_timestamps[min_score_div_idx]
                        region = Segment(
                            start - self.pad_onset, min_score_t + self.pad_offset
                        )
                        active[region, k] = label
                        start = curr_timestamps[min_score_div_idx]
                        curr_scores = curr_scores[min_score_div_idx + 1 :]
                        curr_timestamps = curr_timestamps[min_score_div_idx + 1 :]
                    # switching from active to inactive
                    elif y < self.offset:
                        region = Segment(start - self.pad_onset, t + self.pad_offset)
                        active[region, k] = label
                        start = t
                        is_active = False
                        curr_scores = []
                        curr_timestamps = []
                    curr_scores.append(y)
                    curr_timestamps.append(t)
                # currently inactive
                else:
                    # switching from inactive to active
                    if y > self.onset:
                        start = t
                        is_active = True

            # if active at the end, add final region
            if is_active:
                region = Segment(start - self.pad_onset, t + self.pad_offset)
                active[region, k] = label

        return active

    def __call__(self, scores: SlidingWindowFeature) -> Annotation:
        """Binarize detection scores.

        Args:
            scores (SlidingWindowFeature): Detection scores.

        Returns:
            active (Annotation): Binarized scores.
        """
        active = self.__get_active_regions(scores)
        # because of padding, some active regions might be overlapping: merge them.
        # also: fill same speaker gaps shorter than min_duration_off
        if self.pad_offset > 0.0 or self.pad_onset > 0.0 or self.min_duration_off > 0.0:
            if self.max_duration < float("inf"):
                raise NotImplementedError("This would break current max_duration param")
            active = active.support(collar=self.min_duration_off)

        # remove tracks shorter than min_duration_on
        if self.min_duration_on > 0:
            for segment, track in list(active.itertracks()):
                if segment.duration < self.min_duration_on:
                    del active[segment, track]

        return active



# ruff: noqa: S101, S113
import json
import time
from collections.abc import AsyncGenerator
from typing import Annotated, TypedDict

import requests
from pydantic import Field
from ray import serve

from aana.api.api_generation import Endpoint
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment


@serve.deployment
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        return {"text": [t.lower() for t in text]}


TextList = Annotated[list[str], Field(description="List of text to lowercase.")]


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: list[str]


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(self, text: TextList) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = await self.lowercase_handle.lower(text=text)
        return {"text": lowercase_output["text"]}


class LowercaseSteamEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(
        self, text: TextList
    ) -> AsyncGenerator[LowercaseEndpointOutput, None]:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = await self.lowercase_handle.lower(text=text)
        for t in lowercase_output["text"]:
            yield {"text": t}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    },
]


endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    },
    {
        "name": "lowercase_stream",
        "path": "/lowercase_stream",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseSteamEndpoint,
    },
]


def test_task_queue(create_app):  # noqa: C901
    """Test the Ray Serve app."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port
    route_prefix = ""

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}{route_prefix}/api/ready")
    assert response.status_code == 200
    assert response.json() == {"ready": True}

    # Test lowercase endpoint
    data = {"text": ["Hello World!", "This is a test."]}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200
    lowercase_text = response.json().get("text")
    assert lowercase_text == ["hello world!", "this is a test."]

    # Defer endpoint execution
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase?defer=True",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200
    task_id = response.json().get("task_id")

    # Check the task status with timeout of 10 seconds
    start_time = time.time()
    while time.time() - start_time < 10:
        response = requests.get(
            f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
        )
        task_status = response.json().get("status")
        result = response.json().get("result")
        if task_status == "completed":
            break
        time.sleep(0.1)

    assert task_status == "completed"
    assert result == {"text": ["hello world!", "this is a test."]}

    # Delete the task
    response = requests.get(
        f"http://localhost:{port}{route_prefix}/tasks/delete/{task_id}"
    )
    assert response.status_code == 200
    assert response.json().get("task_id") == task_id

    # Check that the task is deleted
    response = requests.get(
        f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
    )
    assert response.status_code == 404
    assert response.json().get("error") == "NotFoundException"

    # Check non-existent task
    task_id = "d1b1b1b1-1b1b-1b1b-1b1b-1b1b1b1b1b1b"
    response = requests.get(
        f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
    )
    assert response.status_code == 404
    assert response.json().get("error") == "NotFoundException"

    # Test lowercase streaming endpoint
    data = {"text": ["Hello World!", "This is a test."]}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase_stream",
        data={"body": json.dumps(data)},
        stream=True,
    )
    assert response.status_code == 200

    for i, chunk in enumerate(response.iter_content(chunk_size=None)):
        json_data = json.loads(chunk)
        assert "text" in json_data
        assert json_data["text"] == lowercase_text[i]

    # Test task queue with streaming endpoint
    data = {"text": ["Hello World!", "This is a test."]}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase_stream?defer=True",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200
    task_id = response.json().get("task_id")

    # Check the task status with timeout of 10 seconds
    start_time = time.time()
    while time.time() - start_time < 10:
        response = requests.get(
            f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
        )
        task_status = response.json().get("status")
        result = response.json().get("result")
        if task_status == "completed":
            break
        time.sleep(0.1)

    assert task_status == "completed"
    assert [chunk["text"] for chunk in result] == lowercase_text

    # Send 30 tasks to the task queue
    task_ids = []
    for i in range(30):
        data = {"text": [f"Task {i}"]}
        response = requests.post(
            f"http://localhost:{port}{route_prefix}/lowercase_stream?defer=True",
            data={"body": json.dumps(data)},
        )
        assert response.status_code == 200
        task_ids.append(response.json().get("task_id"))

    # Check the task statusES with timeout of 10 seconds
    start_time = time.time()
    completed_tasks = []
    while time.time() - start_time < 10:
        for task_id in task_ids:
            if task_id in completed_tasks:
                continue
            response = requests.get(
                f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
            )
            task_status = response.json().get("status")
            result = response.json().get("result")
            if task_status == "completed":
                completed_tasks.append(task_id)

        if len(completed_tasks) == len(task_ids):
            break
        time.sleep(0.1)

    # Check that all tasks are completed
    for task_id in task_ids:
        response = requests.get(
            f"http://localhost:{port}{route_prefix}/tasks/get/{task_id}"
        )
        response = response.json()
        task_status = response.get("status")
        assert task_status == "completed", response



from aana.core.models.audio import Audio
from aana.core.models.image import Image, ImageInput, ImageInputList
from aana.core.models.media import Media, MediaId
from aana.core.models.video import Video, VideoInput, VideoInputList

__all__ = [
    "MediaId",
    "Audio",
    "Video",
    "Image",
    "Media",
    "ImageInput",
    "VideoInput",
    "ImageInputList",
    "VideoInputList",
]



from typing import Any, TypeVar

from pydantic import BaseModel, RootModel


class BaseListModel(RootModel[list]):
    """The base model for pydantic models with list as root.

    It makes pydantic models with list as root behave like normal lists.
    """

    def __iter__(self):
        """Get iterator for model."""
        return iter(self.root)

    def __len__(self):
        """Get length of model."""
        return len(self.root)

    def __getitem__(self, index):
        """Get item at index of model."""
        return self.root[index]

    def __setitem__(self, index, value):
        """Set item at index of model."""
        self.root[index] = value

    def __delitem__(self, index):
        """Remove item at index of model."""
        del self.root[index]

    def __contains__(self, item):
        """Check if model contains item."""
        return item in self.root

    def __add__(self, other):
        """Add two models."""
        return self.__class__(root=self.root + other.root)


OptionType = TypeVar("OptionType", bound=BaseModel)


def merged_options(default_options: OptionType, options: OptionType) -> OptionType:
    """Merge options into default_options.

    Args:
        default_options (OptionType): default options
        options (OptionType): options to be merged into default_options

    Returns:
        OptionType: merged options
    """
    # if options is None, return default_options
    if options is None:
        return default_options.model_copy()
    # options and default_options have to be of the same type
    if type(default_options) != type(options):
        raise ValueError("Option type mismatch.")  # noqa: TRY003
    default_options_dict = default_options.model_dump()
    for k, v in options.model_dump(exclude_unset=True).items():
        if v is not None:
            default_options_dict[k] = v
    return options.__class__.model_validate(default_options_dict)


def pydantic_to_dict(data: Any) -> Any:
    """Convert all Pydantic objects in the structured data.

    Args:
        data (Any): the structured data

    Returns:
        Any: the same structured data with Pydantic objects converted to dictionaries
    """
    if isinstance(data, BaseModel):
        return data.model_dump()
    elif isinstance(data, list):
        return [pydantic_to_dict(item) for item in data]
    elif isinstance(data, dict):
        return {key: pydantic_to_dict(value) for key, value in data.items()}
    else:
        return data  # return as is for non-Pydantic types


# Workaround for the annoying issue with Pydantic's protected fields.
# See: https://github.com/pydantic/pydantic/issues/9177
# TODO: Remove this workaround once the issue is resolved.
pydantic_protected_fields = [
    "model_computed_fields",
    "model_config",
    "model_construct",
    "model_copy",
    "model_dump",
    "model_dump_json",
    "model_extra",
    "model_fields",
    "model_fields_set",
    "model_json_schema",
    "model_parametrized_name",
    "model_post_init",
    "model_rebuild",
    "model_validate",
    "model_validate_json",
    "model_validate_strings",
]



from pydantic import BaseModel, ConfigDict, Field, field_validator

__all__ = ["SamplingParams"]


class SamplingParams(BaseModel):
    """A model for sampling parameters of LLM.

    Attributes:
        temperature (float): Float that controls the randomness of the sampling. Lower
            values make the model more deterministic, while higher values make
            the model more random. Zero means greedy sampling.
        top_p (float): Float that controls the cumulative probability of the top tokens
            to consider. Must be in (0, 1]. Set to 1 to consider all tokens.
        top_k (int): Integer that controls the number of top tokens to consider. Set
            to -1 to consider all tokens.
        max_tokens (int): The maximum number of tokens to generate.
        repetition_penalty (float): Float that penalizes new tokens based on whether
            they appear in the prompt and the generated text so far. Values > 1 encourage
            the model to use new tokens, while values < 1 encourage the model to repeat
            tokens. Default is 1.0 (no penalty).
        kwargs (dict): Extra keyword arguments to pass as sampling parameters.
    """

    temperature: float = Field(
        default=1.0,
        ge=0.0,
        description=(
            "Float that controls the randomness of the sampling. "
            "Lower values make the model more deterministic, "
            "while higher values make the model more random. "
            "Zero means greedy sampling."
        ),
    )
    top_p: float = Field(
        default=1.0,
        gt=0.0,
        le=1.0,
        description=(
            "Float that controls the cumulative probability of the top tokens to consider. "
            "Must be in (0, 1]. Set to 1 to consider all tokens."
        ),
    )
    top_k: int | None = Field(
        default=None,
        description=(
            "Integer that controls the number of top tokens to consider. "
            "Set to -1 to consider all tokens."
        ),
    )
    max_tokens: int | None = Field(
        default=None, ge=1, description="The maximum number of tokens to generate."
    )
    json_schema: str | None = Field(
        default=None, description="The schema to use for generation."
    )
    regex_string: str | None = Field(
        default=None, description="The regex to use for generation."
    )
    repetition_penalty: float = Field(
        default=1.0,
        description=(
            "Float that penalizes new tokens based on whether they appear in the "
            "prompt and the generated text so far. Values > 1 encourage the model "
            "to use new tokens, while values < 1 encourage the model to repeat tokens. "
            "Default is 1.0 (no penalty)."
        ),
    )
    kwargs: dict = Field(
        default_factory=dict,
        description="Extra keyword arguments to pass as sampling parameters.",
    )

    @field_validator("top_k")
    def check_top_k(cls, v: int):
        """Validates a top_k argument.

        Makes sure it is either -1, or at least 1.

        Args:
            v (int): Value to validate.

        Raises:
            ValueError: The value is not valid.

        Returns:
            The top_k value.
        """
        if v is None:
            return v
        if v < -1 or v == 0:
            raise ValueError(f"top_k must be -1 (disable), or at least 1, got {v}.")  # noqa: TRY003
        return v

    model_config = ConfigDict(
        json_schema_extra={"description": "Sampling parameters for generating text."}
    )



import contextlib
import sys

import click

from aana.exceptions.runtime import DeploymentException
from aana.sdk import AanaSDK
from aana.utils.core import import_from_path


@click.group()
def cli():
    """Aana CLI.

    It provides commands to deploy and build AanaSDK applications.
    """
    pass


def load_app(app_path: str):
    """Load the AanaSDK app from the given path.

    Args:
        app_path (str): Path to the application (module:app).
    """
    sys.path.insert(0, ".")  # to be able import the app from the current directory
    try:
        aana_app = import_from_path(app_path)
    except Exception as e:
        raise ImportError(f"Could not import {app_path}") from e  # noqa: TRY003

    if not isinstance(aana_app, AanaSDK):
        raise TypeError(f"{app_path} is not an AanaSDK instance")  # noqa: TRY003

    return aana_app


@cli.command()
@click.argument("app_path", type=str)
@click.option(
    "--host", default="127.0.0.1", type=str, help="Host address (default: 127.0.0.1)"
)
@click.option(
    "--port", default=8000, type=int, help="Port to run the application (default: 8000)"
)
@click.option("--hide-logs", is_flag=True, help="Hide logs")
@click.option(
    "--ray-address",
    default="auto",
    type=str,
    help="Address of the Ray cluster (default: auto)",
)
@click.option(
    "--ray-dashboard-host",
    default="127.0.0.1",
    type=str,
    help=(
        "The host to bind the dashboard server to. Can either be "
        "localhost (127.0.0.1) or 0.0.0.0 (available from all interfaces). "
        "By default, this is set to localhost to prevent access from external machines."
    ),
)
@click.option(
    "--ray-dashboard-port",
    default=8265,
    type=int,
    help="The port to bind the dashboard server to (default: 8265)",
)
@click.option(
    "--skip-migrations",
    is_flag=True,
    help="Skip migrations before deploying",
)
@click.option(
    "--sequential-deploy",
    is_flag=True,
    help="Deploy the application's deployments sequentially",
)
def deploy(
    app_path: str,
    host: str,
    port: int,
    hide_logs: bool,
    ray_address: str,
    ray_dashboard_host: str,
    ray_dashboard_port: int,
    skip_migrations: bool,
    sequential_deploy: bool,
):
    """Deploy the application.

    APP_PATH: Path to the application (module:app).
    """
    aana_app = load_app(app_path)
    if not skip_migrations:
        aana_app.migrate()
    show_logs = not hide_logs
    aana_app.connect(
        port=port,
        host=host,
        show_logs=show_logs,
        address=ray_address,
        dashboard_host=ray_dashboard_host,
        dashboard_port=ray_dashboard_port,
    )
    with contextlib.suppress(
        DeploymentException
    ):  # we have a nice error message for this
        aana_app.deploy(blocking=True, sequential=sequential_deploy)


@cli.command()
@click.argument("app_path", type=str)
@click.option(
    "--host",
    default="0.0.0.0",  # noqa: S104
    type=str,
    help="Host address (default: 0.0.0.0)",
)
@click.option(
    "--port", default=8000, type=int, help="Port to run the application (default: 8000)"
)
@click.option(
    "--app-config-name",
    default="app_config",
    type=str,
    help="App config name (default: app_config)",
)
@click.option(
    "--config-name",
    default="config",
    type=str,
    help="Config name (default: config)",
)
def build(app_path: str, host: str, port: int, app_config_name: str, config_name: str):
    """Build the application.

    APP_PATH: Path to the application (module:app).
    """
    aana_app = load_app(app_path)
    aana_app.build(
        import_path=app_path,
        host=host,
        port=port,
        app_config_name=app_config_name,
        config_name=config_name,
    )


@cli.command()
@click.argument("app_path", type=str)
def migrate(app_path: str):
    """Migrate the database for the application.

    APP_PATH: Path to the application (module:app).
    """
    aana_app = load_app(app_path)
    aana_app.migrate()



# ruff: noqa: S101
from importlib import resources
from pathlib import Path

import pytest

from aana.core.models.audio import Audio
from aana.core.models.video import Video, VideoInput
from aana.integrations.external.yt_dlp import download_video
from aana.processors.video import extract_audio


@pytest.mark.parametrize("audio_file", ["squirrel.wav", "physicsworks.wav"])
def test_audio(audio_file):
    """Test that the audio can be created from path, url(pending), or content."""
    # Test creation from a path
    try:
        path = resources.files("aana.tests.files.audios") / audio_file
        audio = Audio(path=path, save_on_disk=False)
        assert audio.path == path
        assert audio.content is None
        assert audio.url is None

        content = audio.get_content()
        assert len(content) > 0
    finally:
        audio.cleanup()

    # Test creation from URL
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        audio = Audio(url=url, save_on_disk=False)
        assert audio.path is None
        assert audio.content is None
        assert audio.url == url

    finally:
        audio.cleanup()

    # Test creation from content
    try:
        path = resources.files("aana.tests.files.audios") / audio_file
        content = path.read_bytes()
        audio = Audio(content=content, save_on_disk=False)
        assert audio.path is None
        assert audio.content == content
        assert audio.url is None

        assert audio.get_content() == content
    finally:
        audio.cleanup()


def test_audio_path_not_exist():
    """Test that the audio can't be created from path if the path doesn't exist."""
    path = Path("path/to/audio_that_does_not_exist.wav")
    with pytest.raises(FileNotFoundError):
        Audio(path=path)


@pytest.mark.parametrize("audio_file", ["squirrel.wav", "physicsworks.wav"])
def test_save_audio(audio_file):
    """Test that save_on_disk works for audio."""
    # Test that the audio is saved to disk when save_on_disk is True
    try:
        path = resources.files("aana.tests.files.audios") / audio_file
        audio = Audio(path=path, save_on_disk=True)
        assert audio.path == path
        assert audio.content is None
        assert audio.url is None
        assert audio.path.exists()
    finally:
        audio.cleanup()
    assert audio.path.exists()  # Cleanup should NOT delete the file if path is provided

    # Test saving from audio URL to disk and read
    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.wav"

        audio = Audio(url=url, save_on_disk=True)
        assert audio.content is None
        assert audio.url is not None
    finally:
        audio.cleanup()


@pytest.mark.parametrize("audio_file", ["squirrel.wav", "physicsworks.wav"])
def test_cleanup(audio_file):
    """Test that cleanup works for audios."""
    try:
        path = resources.files("aana.tests.files.audios") / audio_file
        audio = Audio(path=path, save_on_disk=True)
        assert audio.path.exists()
    finally:
        audio.cleanup()
        assert (
            audio.path.exists()
        )  # Cleanup should NOT delete the file if path is provided


def test_at_least_one_input():
    """Test that at least one input is provided for audio."""
    with pytest.raises(ValueError):
        Audio(save_on_disk=False)

    with pytest.raises(ValueError):
        Audio(save_on_disk=True)


def test_extract_audio():
    """Test download_video and extract_audio."""
    # Test VideoInput with video path (tests download_video and extract_audio): return audio bytes
    path = resources.files("aana.tests.files.videos") / "physicsworks.webm"
    video_input = VideoInput(path=str(path))
    video = download_video(video_input)
    assert isinstance(video, Video)
    assert video.path == path
    assert video.content is None
    assert video.url is None

    audio = extract_audio(video)
    assert isinstance(audio, Audio)
    assert audio.path is not None
    assert audio.content is not None
    assert audio.url is None

    # Test audio from VideoInput (for audio input): return audio bytes
    path = resources.files("aana.tests.files.audios") / "physicsworks.wav"
    video_input = VideoInput(path=str(path))
    video = download_video(video_input)
    assert isinstance(video, Video)
    audio = extract_audio(video)
    assert isinstance(audio, Audio)
    assert audio.content is not None
    assert audio.path is not None
    assert audio.url is None

    # Test audio from VideoInput (for no audio channel): return empty bytes
    path = resources.files("aana.tests.files.videos") / "squirrel_no_audio.mp4"
    video_input = VideoInput(path=str(path))
    video = download_video(video_input)
    audio = extract_audio(video)
    assert isinstance(audio, Audio)
    assert audio.content == b""
    assert audio.url is None
    assert audio.path is not None

    try:
        url = "https://mobius-public.s3.eu-west-1.amazonaws.com/squirrel.mp4"
        video_input = VideoInput(url=url)
        video = download_video(video_input)
        audio = extract_audio(video)
        assert isinstance(audio, Audio)
        assert audio.path is not None
        assert audio.content is not None
        assert audio.path.exists()
        assert video.media_id == video_input.media_id
        assert audio.media_id == f"audio_{video.media_id}"

    finally:
        video.cleanup()
        audio.cleanup()

    # Test loading numpy from the Audio object
    path = resources.files("aana.tests.files.audios") / "physicsworks.wav"
    audio = Audio(path=path, save_on_disk=True)
    assert audio.get_numpy() is not None

    path = resources.files("aana.tests.files.audios") / "squirrel.wav"
    audio = Audio(path=path, save_on_disk=True)
    assert audio.get_numpy().all() == 0

    # Test content of Audio objects created in different ways from same source video is the same.
    video_path = resources.files("aana.tests.files.videos") / "physicsworks.webm"
    video = Video(path=video_path)
    extracted_audio_1 = extract_audio(video)

    video_input = VideoInput(path=str(video_path))
    downloaded_video = download_video(video_input)
    extracted_audio_2 = extract_audio(downloaded_video)

    # Paths will be different since Audio object is created from two different Video objects,
    # but the audio content should be the same
    assert extracted_audio_1.path != extracted_audio_2.path
    assert (extracted_audio_1.get_numpy().all()) == extracted_audio_2.get_numpy().all()
